from typing import List, Optional

from open_mafia_engine.core.all import Actor, GameObject

from .kills import LynchAction
from .voting import Tally


class LynchTally(Tally):
    """Vote tally that lynches the vote leader."""

    def respond_leader(self, leader: GameObject) -> Optional[List[LynchAction]]:
        """Override this for particular behavior."""
        if isinstance(leader, Actor):
            return [LynchAction(self.game, self, target=leader)]



from open_mafia_engine.builders.for_testing import make_test_game
from open_mafia_engine.built_in.voting import Tally
from open_mafia_engine.core.event_system import Action
from open_mafia_engine.core.phase_cycle import PhaseChangeAction
from open_mafia_engine.core.state import Ability, EActivate


def test_activation_by_event():
    """Tests activation via events, directly or via converters."""

    game = make_test_game(["Alice", "Bob"])
    alice = game.actors[0]
    bob = game.actors[1]

    a_v = alice.abilities[0]  # VoteAbility(game, owner=alice, name="Vote")
    b_v = bob.abilities[0]  # VoteAbility(game, owner=bob, name="Vote")

    game.phase_system.bump_phase()  # start the day

    tally: Tally = game.aux.filter_by_type(Tally)[0]
    game.process_event(EActivate(game, a_v, target=bob), process_now=True)
    assert tally.results.vote_leaders == [bob]
    game.process_event(
        EActivate(game, "Alice/ability/Vote", target="unvote"), process_now=True
    )
    assert tally.results.vote_leaders == []


def test_phase_constraint():
    """Tests phase constraint for voting."""

    game = make_test_game(["Alice", "Bob"])
    alice = game.actors[0]
    bob = game.actors[1]

    a_v = alice.abilities[0]  # VoteAbility(game, owner=alice, name="Vote")
    b_v = bob.abilities[0]  # VoteAbility(game, owner=bob, name="Vote")

    tally: Tally = game.aux.filter_by_type(Tally)[0]

    # Won't work - game still in "startup"
    game.process_event(EActivate(game, a_v, target=bob), process_now=True)
    assert tally.results.vote_leaders == []

    game.phase_system.bump_phase()  # start the day

    # Now it should work :)
    game.process_event(EActivate(game, a_v, target=bob), process_now=True)
    assert tally.results.vote_leaders == [bob]


def test_alive_constraint():
    """Tests alive constraint."""

    game = make_test_game(["Alice", "Bob", "Charlie"])
    alice = game.actors[0]
    bob = game.actors[1]

    # a_v = alice.abilities[0]  # VoteAbility(game, owner=alice, name="Vote")
    b_v = bob.abilities[0]  # VoteAbility(game, owner=bob, name="Vote")

    tally: Tally = game.aux.filter_by_type(Tally)[0]

    game.phase_system.bump_phase()  # start the day

    # Test whether alive constraint works :)
    bob.status["dead"] = True
    e = EActivate(game, b_v, target=alice)  # should fail to activate - Bob is dead
    game.process_event(e, process_now=True)
    assert tally.results.vote_counts == []


def test_gen_action_ability():
    """Tests action and ability generation."""

    actor_names = ["Alpha", "Bravo"]

    def fake_action(self: Action):
        assert self.game.actor_names == actor_names

    AcFake = Action.generate(fake_action, name="AcFake", doc="Fake action")
    AbFake = Ability.generate(AcFake, name="AbFake", doc="Fake ability")

    @Ability.generate
    def AbFake2(self: Action):
        """fake ability 2"""
        assert self.game.actor_names == actor_names

    game = make_test_game(actor_names)

    a_abil = AbFake(game, owner=game.actors[0], name="a_abil")
    a_abil.activate()

    b_abil = AbFake2(game, owner="Bravo", name="b_abil")
    b_abil.activate()


import warnings
from collections.abc import MutableMapping
from typing import Dict, Generic, TypeVar

with warnings.catch_warnings():
    warnings.simplefilter("ignore")

    from fuzzywuzzy import process  # Remove warning about performance

T = TypeVar("T")


class Matcher(MutableMapping, Generic[T]):
    """Base class for target matching.

    The base version only does exact matches, and throws KeyError if not found.
    (Essentially, just a `Dict[str, T]`)
    """

    def __init__(self, choices: Dict[str, T]):
        #
        self._choices: Dict[str, T] = {}
        # Make sure to check
        for k, v in choices.items():
            self[k] = v

    def match(self, query: str) -> T:
        """Tries to match `query` to your object."""
        return self._choices[query]

    @property
    def choices(self) -> Dict[str, T]:
        return dict(self._choices)

    @choices.setter
    def choices(self, v: Dict[str, T]):
        self._choices = dict(v)

    def __getitem__(self, key: str) -> T:
        if key in self._choices:
            return self._choices[key]
        return self.match(key)

    def __setitem__(self, key: str, value: T):
        self._choices[key] = value

    def __delitem__(self, key: str):
        del self._choices[key]

    def __iter__(self):
        return iter(self.choices)  # making a copy

    def __len__(self):
        return len(self._choices)


class MatcherLowercase(Matcher, Generic[T]):
    """Matches everything by `query.lower()`."""

    def __setitem__(self, key: str, value: T):
        self._choices[key.lower()] = value

    def match(self, query: str) -> T:
        return self._choices[query.lower()]


class FuzzyMatcher(Matcher, Generic[T]):
    """Fuzzy matcher, with a score cutoff, using `fuzzywuzzy`.

    Parameters
    ----------
    choices : Dict[str, T]
        Mapping of name to whatever you want.
    score_cutoff : int
        The minimum score to use.
        Default is 0, in which case we use the best match, even if it's bad.
        If set higher, `match()` may raise KeyError.
    use_lower : bool
        Whether to use str.lower() for keys and queries.
        This probably improves matching in real-world cases.
        Default is True.
    """

    def __init__(
        self,
        choices: Dict[str, T] = None,
        *,
        score_cutoff: int = 0,
        use_lower: bool = True
    ):
        self.score_cutoff = int(score_cutoff)
        self.use_lower = bool(use_lower)
        super().__init__(choices)

    def __setitem__(self, key: str, value: T):
        if self.use_lower:
            key = key.lower()
        self._choices[key] = value

    def match(self, query: str) -> T:
        if self.use_lower:
            query = query.lower()
        options = list(self._choices.keys())
        # see: https://github.com/seatgeek/fuzzywuzzy
        res = process.extractOne(query, options, score_cutoff=self.score_cutoff)
        if res is None:
            raise KeyError(query)
        return self._choices[res[0]]



from collections.abc import Mapping, Sequence
from typing import List

from open_mafia_engine.util.repr import ReprMixin


class NamedList(Mapping, Sequence, ReprMixin):
    """Named list: works like a list and a dict based on the key."""

    def __init__(self, inner: List, attr: str = "name"):
        self._inner = list(inner)
        self._attrs = {getattr(o, attr): o for o in self._inner}
        self._attr = attr

    @property
    def inner(self) -> List:
        return self._inner

    @property
    def attr(self) -> str:
        return self._attr

    def __str__(self) -> str:
        return str(self._inner)

    def __getitem__(self, k):
        if isinstance(k, int):
            return self._inner[k]
        return self._attrs[k]

    def __len__(self) -> int:
        return len(self._inner)

    def __iter__(self):
        return iter(self._inner)


# flake8: noqa

from open_mafia_engine.core.enums import ActionResolutionType, Outcome

from .auxiliary import AuxHelper, AuxObject, RemoveAuxAction
from .builder import GameBuilder, game_builder
from .converters import (
    get_ability_by_path,
    get_actor_by_name,
    get_faction_by_name,
    get_phase_by_name,
    get_trigger_by_path,
)
from .ender import EGameEnded, EndTheGame, GameEnder
from .event_system import (
    Action,
    ActionInspector,
    ActionQueue,
    CancelAction,
    ConditionalCancelAction,
    Constraint,
    EPostAction,
    EPreAction,
    Event,
    EventEngine,
    EventHandler,
    Subscriber,
    _HandlerFunc,
    handler,
    handles,
)
from .game import Game
from .game_object import GameObject, converter, inject_converters
from .naming import ABILITY, PATH_SEP, TRIGGER, get_parts, get_path
from .outcome import EOutcomeAchieved, OutcomeAction
from .phase_cycle import (
    AbstractPhaseSystem,
    ActionResolutionType,
    ETryPhaseChange,
    EPrePhaseChange,
    EPostPhaseChange,
    Phase,
    PhaseChangeAction,
    SimplePhaseCycle,
)
from .state import (
    Ability,
    Actor,
    ATBase,
    ATConstraint,
    ConstraintActorTargetsAlive,
    ConstraintOwnerAlive,
    EActivate,
    EStatusChange,
    Faction,
    OutcomeChecker,
    Status,
    Trigger,
)



from typing import List

from open_mafia_engine.builders.for_testing import make_test_game
from open_mafia_engine.core.event_system import (
    Action,
    Event,
    Subscriber,
    handler,
    handles,
)
from open_mafia_engine.core.state import EStatusChange


def test_events_and_subscribers():
    """General test of events and subscribers.

    Also makes sure order is deterministic.
    """

    actor_names = ["Alpha", "Bravo"]
    game = make_test_game(actor_names)

    log = []

    class EFake(Event):
        """Fake event"""

    class X1(Action):
        def doit(self):
            log.append("X1")

    class A(Subscriber):
        @handles(Event)
        def f(self, event: Event):
            log.append("A.f")

    class B(A):
        @handler
        def g(self, event: EFake) -> None:
            log.append("B.g")

    class C(Subscriber):
        @handles(EFake)
        def h1(self, event) -> List[X1]:
            log.append("C.h1")
            return [X1(self.game, self)]

        @handler
        def h2(self, event: X1.Pre) -> List[Action]:
            log.append("C.h2")
            return []

        @handler
        def h3(self, event: X1.Post) -> None:
            log.append("C.h3")

        # Also legal:
        # -> Union[List[Action], List[X1]]

        # Illegal handlers

        # @handles(Event)
        # def z1(self, event) -> Action:
        #     return Action(self.game, self)

        # @handles(Event)
        # def z2(self, event) -> Optional[Action]:
        #     return None

    a = A(game)
    b = B(game)
    c = C(game)

    e = EFake(game)
    game.process_event(e, process_now=True)

    assumed_log = [
        "B.g",
        "C.h1",
        "A.f",
        "A.f",
        "C.h2",
        "A.f",
        "A.f",
        "X1",
        "C.h3",
        "A.f",
        "A.f",
    ]
    assert log == assumed_log


def test_status_event():
    """Tests status change events."""

    actor_names = ["Alpha", "Bravo"]
    game = make_test_game(actor_names)

    log = []

    class Chk(Subscriber):
        @handler
        def f(self, event: EStatusChange):
            log.append(event.new_val)

    Chk(game)

    game.actors[0].status["key1"] = 2

    assert log == [2]



from typing import List

from open_mafia_engine.built_in.all import *
from open_mafia_engine.core.all import *


@game_builder("test")
def make_test_game(player_names: List[str], n_mafia: int = 1) -> Game:
    """Game builder for testing purposes.

    Always assigns mafia, then non-mafia, in player name order.
    Must be at least 2 players (1 mafia, 1 town).

    Parameters
    ----------
    n_mafia : int = 1
        How many mafia players to add.
    """

    n_mafia = int(n_mafia)

    n = len(player_names)
    assert isinstance(n_mafia, int)
    assert n_mafia > 0
    assert n > n_mafia

    game = Game()
    mafia = Faction(game, "Mafia")
    OCLastFactionStanding(game, mafia)
    FactionChatCreatorAux(game, mafia)

    town = Faction(game, "Town")
    OCLastFactionStanding(game, town)

    # Aux objects
    GameEnder(game)  # ends the game when all factions get an Outcome
    tally = LynchTally(game)  # lynch tally

    n_town = n - n_mafia

    vote_desc = "Vote for your target. This is a free action."

    for i in range(n_mafia):
        act = Actor(game, player_names[i])
        mafia.add_actor(act)
        # Voting
        vote = VoteAbility(game, act, name="Vote", tally=tally, desc=vote_desc)
        PhaseConstraint(game, vote, phase="day")
        # Mafia kill
        mk = KillAbility(
            game,
            act,
            name="Mafia Kill",
            desc="Kill the target. Only 1 mafioso can use this.",
        )
        LimitPerPhaseActorConstraint(game, mk, limit=1)
        LimitPerPhaseKeyConstraint(game, mk, key="mafia_kill_limit")
        PhaseConstraint(game, mk, phase="night")
        ConstraintNoSelfFactionTarget(game, mk)
        if i == 2:
            # Second mafioso can roleblock
            block = RoleBlockAbility(
                game,
                act,
                name="Roleblock",
                desc="Blocks the target from acting this night.",
            )
            PhaseConstraint(game, block, phase="night")
            LimitPerPhaseActorConstraint(game, block, limit=1)
            ConstraintNoSelfFactionTarget(game, block)

    for i in range(n_town):
        act = Actor(game, player_names[n_mafia + i])
        town.add_actor(act)
        # Voting
        vote = VoteAbility(game, act, name="Vote", tally=tally, desc=vote_desc)
        PhaseConstraint(game, vote, phase="day")
        if i == 1:
            # Second townie is a protector/doctor
            prot = ProtectFromKillAbility(game, act, name="Protect")
            PhaseConstraint(game, prot, phase="night")
            LimitPerPhaseActorConstraint(game, prot, limit=1)
            ConstraintNoSelfTarget(game, prot)
        elif i == 2:
            # Third townie is a detective
            insp = InspectFactionAbility(game, act, name="Faction Inspect")
            LimitPerPhaseActorConstraint(game, insp, limit=1)
            PhaseConstraint(game, insp, phase="night")
            ConstraintNoSelfTarget(game, insp)
            pass
        elif i == 3:
            # Fourth townie is a redirector
            redir = CreateRedirectAbility(game, act, name="Redirect")
            LimitPerPhaseActorConstraint(game, redir, limit=1)
            PhaseConstraint(game, redir, phase="night")
        # TODO: Other abilities

    return game



from dataclasses import dataclass, field
from typing import Any, Dict, Tuple, TypeVar

from open_mafia_engine.errors import MafiaError

__all__ = ["TUser", "RawCommand", "MafiaBadCommand"]

TUser = TypeVar("TUser")  # User type in your application


@dataclass(frozen=True)
class RawCommand:
    """Raw parsed command"""

    source: str
    name: str
    args: Tuple = tuple()
    kwargs: Dict[str, Any] = field(default_factory=dict)


class MafiaBadCommand(MafiaError, KeyError):
    """Bad command was passed."""

    def __init__(self, rc: RawCommand) -> None:
        self.raw_command = rc
        super().__init__(f"Bad command: {rc!r}")


from __future__ import annotations

import inspect
import logging
import warnings
from abc import abstractmethod
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    DefaultDict,
    Dict,
    List,
    Optional,
    Type,
    Union,
    get_args,
    get_origin,
    get_type_hints,
)

from makefun import partial, wraps
from sortedcontainers import SortedList

from open_mafia_engine.core.game_object import GameObject
from open_mafia_engine.errors import MafiaBadHandler

if TYPE_CHECKING:
    from open_mafia_engine.core.game import Game

NoneType = type(None)

logger = logging.getLogger(__name__)


class Event(GameObject):
    """Core event object."""

    def __init__(self, game, /):
        super().__init__(game)


class _ActionEvent(Event):
    """Base class for pre- and post-action events."""

    def __init__(self, game, action: Action, /):
        self._action = action
        super().__init__(game)

    @property
    def action(self) -> Action:
        return self._action


class EPreAction(_ActionEvent):
    """Pre-action event."""


class EPostAction(_ActionEvent):
    """Post-action event."""


class Action(GameObject):
    """Core action object.

    Attributes
    ----------
    game : Game
    source : GameObject
        The object that generated this action.
    priority : float
    canceled : bool

    Abstract Methods
    ----------------
    doit(self) -> None
        Performs the action.

    Class Attributes
    ----------------
    Pre : Type[EPreAction]
    Post : Type[EPostAction]
        Pre- and post-action event classes to use with `action.pre` and `action.post`.
        You may override these with your own when subclassing.
    """

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        *,
        priority: float = 0.0,
        canceled: bool = False,
    ):
        self._priority = float(priority)
        self._canceled = bool(canceled)
        if not isinstance(source, GameObject):
            raise TypeError(f"Expected GameObject, got {source!r}")
        self._source = source
        super().__init__(game)

    # NOTE: You can override these classes, even inline.
    Pre = EPreAction
    Post = EPostAction

    @abstractmethod
    def doit(self):
        """Performs the action."""

        # Override this.

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        assert issubclass(cls.Pre, EPreAction)
        assert issubclass(cls.Post, EPostAction)

    @property
    def source(self) -> GameObject:
        return self._source

    @property
    def priority(self) -> float:
        return self._priority

    @priority.setter
    def priority(self, v: float):
        v = float(v)
        # You can add an event handler here
        self._priority = v

    @property
    def canceled(self) -> bool:
        return self._canceled

    @canceled.setter
    def canceled(self, v: bool):
        v = bool(v)
        # You can add an event handler here
        self._canceled = v

    def pre(self) -> EPreAction:
        """Create a pre-event for this action."""
        return self.Pre(self.game, self)

    def post(self) -> EPostAction:
        """Create a post-event for this action."""
        return self.Post(self.game, self)

    @classmethod
    def generate(
        cls, func: Callable, name: str = None, doc: str = None
    ) -> Type[Action]:
        """Create an Action subtype from a function.

        Parameters
        ----------
        func : Callable
            Function that does something. :)
        name : str
            What name to use. By default, will auto-generate a name from the func name.
        doc : str
            Docstring. By default, will use func doc, with prepended Action info.
        """

        if name is None:
            # NOTE: We can add random bits at the end to avoid conflicts
            # But this might mess up serialization?
            # from uuid import uuid4
            # rand_ = str(uuid4()).replace("-", "")[-8:]
            name = f"{cls.__name__}_{func.__name__}"

        if doc is None:
            doc = "(GENERATED ACTION) " + (func.__doc__ or "")

        # Get all signatures, parameters to merge
        sig_core = inspect.signature(cls.__init__)
        params_core = list(sig_core.parameters.values())
        sig_func = inspect.signature(func)
        params_func = list(sig_func.parameters.values())

        # Make sure the function has "self" as 0th arg
        if (len(params_func) < 1) or (params_func[0].name != "self"):
            raise TypeError(f"Function requires `self` argument to be first.")
        # Give a hint via warnings :)
        if params_func[0].annotation is None:
            warnings.warn(
                "We suggest you annotate like this to improve code editor experience:\n"
                f"  {func.__name__}(self: Action, ...)"
            )

        # Generate the list of params by merging
        DEFAULTS = {k: v.default for k, v in sig_core.parameters.items()}
        DEFAULTS.update({k: v.default for k, v in sig_func.parameters.items()})
        params_res: List[inspect.Parameter] = []
        attr_names = []
        kinds = [
            inspect.Parameter.POSITIONAL_ONLY,
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.VAR_POSITIONAL,
            inspect.Parameter.KEYWORD_ONLY,
            inspect.Parameter.VAR_KEYWORD,
        ]
        i_core = 0
        i_func = 0
        for kind in kinds:
            while (i_core < len(params_core)) and (params_core[i_core].kind == kind):
                params_res.append(params_core[i_core])
                i_core += 1
            while (i_func < len(params_func)) and (params_func[i_func].kind == kind):
                pfi = params_func[i_func]
                conflict_params = [p for p in params_res if p.name == pfi.name]
                if len(conflict_params) == 0:
                    params_res.append(pfi)
                    attr_names.append(pfi.name)
                elif len(conflict_params) == 1:
                    # conflict_params[0].default = pfi.default, except can't set param!
                    conf = conflict_params[0]
                    replacement = conf.replace(default=pfi.default)
                    params_res[params_res.index(conf)] = replacement
                    # don't add it - will conflict
                i_func += 1
        # Note: we make sure that we don't have duplicate *args, **kwargs
        sig_res = inspect.Signature(params_res)

        @wraps(cls.__init__, new_sig=sig_res)
        def __init__(
            self,
            game,
            /,
            *args,
            priority: float = DEFAULTS.get("priority", 0.0),
            canceled: bool = DEFAULTS.get("canceled", False),
            **kwargs,
        ):
            super(type(self), self).__init__(game, priority=priority, canceled=canceled)

            # Set attributes
            bs = sig_res.bind(
                self, game, *args, priority=priority, canceled=canceled, **kwargs
            )  # FIXME: Should we pass in `game`?
            bs.apply_defaults()
            for attr_name in attr_names:
                setattr(self, attr_name, bs.arguments[attr_name])
            # NOTE: We can't keep the signature, because someone might change
            # the arguments on the class instance; we have to "re-parse" instead

            # Housekeeping
            self._func = func
            self._sig = sig_res
            self._attr_names = tuple(attr_names)

        def doit(self):
            """Performs the action (generated from function)."""

            func = self._func
            sig = self._sig
            attr_names = self._attr_names

            # "re-parse" the signature
            args = []
            kwargs = {}
            for k, p in sig.parameters.items():
                if k not in attr_names:
                    continue
                v = getattr(self, k)
                if p.kind in [
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                ]:
                    args += v
                elif p.kind == inspect.Parameter.VAR_POSITIONAL:
                    args.extend(v)
                elif p.kind == inspect.Parameter.KEYWORD_ONLY:
                    kwargs[k] = v
                elif p.kind == inspect.Parameter.VAR_KEYWORD:
                    kwargs.update(v)

            func(self, *args, **kwargs)

        GeneratedAction = type(
            name, (cls,), {"__init__": __init__, "doit": doit, "__doc__": doc}
        )
        return GeneratedAction


class CancelAction(Action):
    """Action that cancels other actions."""

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        target: Action,
        *,
        priority: float = 50.0,
        canceled: bool = False,
    ):
        self._target = target
        super().__init__(game, source, priority=priority, canceled=canceled)

    @property
    def target(self) -> Action:
        return self._target

    @target.setter
    def target(self, v: Action):
        if not isinstance(v, Action):
            raise TypeError(f"Can only cancel Actions, but got {v!r}")
        self._target = v

    def doit(self):
        self.target.canceled = True


class ConditionalCancelAction(CancelAction):
    """Cancels the action, but checks `condition` just in case again.

    If `condition(action)`, actually does cancel the action.
    """

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        target: Action,
        condition: Callable[[Action], bool],
        *,
        priority: float = -100,
        canceled: bool = False,
    ):
        self._condition = condition
        super().__init__(game, source, target, priority=priority, canceled=canceled)

    @property
    def condition(self) -> Callable[[Action], bool]:
        return self._condition

    def doit(self):
        if self.condition(self.target):
            super().doit()


class ActionInspector(object):
    """Helper to inspect Action objects."""

    def __init__(self, action: Action):
        self._action = action

    @property
    def action(self) -> Action:
        return self._action

    @property
    def ignored_args(self) -> List[str]:
        """Arguments that are ignored"""
        return ["self", "game", "priority", "canceled", "return"]

    @property
    def type_hints(self) -> Dict[str, Type]:
        """Type hints, without ignored arguments."""
        raw = get_type_hints(self.action.__init__)
        return {k: v for k, v in raw.items() if k not in self.ignored_args}

    @property
    def param_names(self) -> List[str]:
        """Parameter names."""
        return list(self.type_hints.keys())

    def params_of_type(self, T: Type) -> List[str]:
        """Returns parameter names that have the given type."""

        def chk(v):
            try:
                return issubclass(v, T)
            except Exception:
                return False

        return [k for k, v in self.type_hints.items() if chk(v)]

    def values_of_type(self, T: Type) -> Dict[str, Any]:
        return {k: self.extract_value(k) for k in self.params_of_type(T)}

    def extract_value(self, param: str) -> Any:
        """Gets value for the parameter."""
        # TODO: Make this smarter? :)
        return getattr(self.action, param)

    def set_value(self, param: str, obj: Any):
        """Sets value for the parameter."""
        # TODO: Make this smarter? :)
        if not hasattr(self.action, param):
            warnings.warn(f"Action has no parameter {param!r}, setting anyways.")
        return setattr(self.action, param, obj)


class ActionQueue(GameObject):
    """Action queue.

    queue : List[Action]
        Actions are sorted by decreasing priority. Ties are broken by insertion order.
    history : List[Action]
        Actions are stored in the order they were processed in, including sub-queues.
    """

    MAX_DEPTH: int = 20

    def __init__(self, game, depth: int = 0):
        depth = int(depth)
        if depth > self.MAX_DEPTH:
            raise RecursionError(f"Reached recursion limit of {self.MAX_DEPTH}")
        self._depth = depth
        self._queue = SortedList([], key=self._action_sorter)
        self._history: List[Action] = []
        super().__init__(game)

    def __len__(self) -> int:
        return len(self._queue)

    @staticmethod
    def _action_sorter(action: Action) -> float:
        """Key function for sorting - by decreasing priority."""
        return -action.priority

    @property
    def depth(self) -> int:
        return self._depth

    @property
    def queue(self) -> List[Action]:
        return list(self._queue)

    @property
    def history(self) -> List[Action]:
        return list(self._history)

    def enqueue(self, action: Action):
        """Add an action to the queue."""
        if not isinstance(action, Action):
            raise TypeError(f"Expected Action, got {action!r}")
        self._queue.add(action)

    def pop_batch(self) -> List[Action]:
        """Gets the next batch of actions, removing them from the queue.

        If there are no more actions, this returns an empty list.
        """
        if len(self._queue) == 0:
            return []

        res = []
        priority = self._queue[0].priority
        while (len(self._queue) > 0) and (self._queue[0].priority == priority):
            action: Action = self._queue.pop(0)
            res.append(action)
        return res

    def peek_batch(self) -> List[Action]:
        """Peek at the next batch of actions, without removing them.

        If there are no more actions, this returns an empty list.
        """
        if len(self._queue) == 0:
            return []
        res = []
        i = 0
        priority = self._queue[0].priority
        while (i < len(self._queue)) and (self._queue[i].priority == priority):
            res.append(self._queue[i])
        return res

    def add_history(self, actions: List[Action]):
        """Adds the actions to history."""
        self._history.extend(actions)

    def __repr__(self) -> str:
        cn = type(self).__qualname__
        return f"<{cn} with {len(self._queue)} queued, {len(self._history)} in history>"

    def process_next_batch(self):
        next_batch: List[Action] = self.pop_batch()

        pre_responses = []
        for action in next_batch:
            pre_responses += self.game.event_engine.broadcast(action.pre())
        pre_queue = ActionQueue(self.game, depth=self._depth + 1)
        for pre_r in pre_responses:
            pre_queue.enqueue(pre_r)
        pre_queue.process_all()
        self.add_history(pre_queue.history)

        # Run the actions themselves
        for action in next_batch:
            if not action.canceled:
                action.doit()
                self.add_history([action])

        # Get and run all post-action responses
        post_responses = []
        for action in next_batch:
            if not action.canceled:
                post_responses += self.game.event_engine.broadcast(action.post())
        post_queue = ActionQueue(self.game, depth=self._depth + 1)
        for post_r in post_responses:
            post_queue.enqueue(post_r)
        post_queue.process_all()
        self.add_history(post_queue.history)

    def process_all(self):
        while len(self) > 0:
            self.process_next_batch()


class Subscriber(GameObject):
    """Base class for objects that listen to events.

    Creating Event Handlers
    -----------------------
    Subclass from this and use a `handler` or `handles` decorator.
    The following result in the same calls behavior:

        class MySub(Subscriber):
            @handles(EPreAction)
            def handler_1(self, event) -> Optional[List[Action]]:
                return None

            @handler
            def handler_2(self, event: EPreAction):
                return []

    Adding Constraints
    ------------------
    `Constraint`s are added after the Subscriber is created.
    """

    def __init__(self, game: Game, /, *, use_default_constraints: bool = True):
        self._use_default_constraints = bool(use_default_constraints)
        super().__init__(game)
        self._constraints: List[Constraint] = []
        self._handler_funcs: List[_HandlerFunc] = []
        self._sub()
        if self._use_default_constraints:
            self.add_default_constraints()

    @property
    def constraints(self) -> List[Constraint]:
        return list(self._constraints)

    @property
    def use_default_constraints(self) -> bool:
        """Whether this object is using default constraints for this class."""
        return self._use_default_constraints

    def _sub(self):
        """Subscribe to events in the current game. This should happen automatically."""
        for handler in self.get_handlers():
            hf = self.game.event_engine.add_handler(handler, self)
            self._handler_funcs.append(hf)

    def _unsub(self):
        """Unsubscrive all handlers from the current game."""
        self.game.event_engine.remove_subscriber(self)

    @classmethod
    def get_handlers(cls) -> List[EventHandler]:
        """Returns all event handlers for this class."""
        res = []
        for T in cls.mro():
            for x in T.__dict__.values():
                if isinstance(x, EventHandler):
                    res.append(x)
        return res

    @property
    def handler_funcs(self) -> List[_HandlerFunc]:
        return list(self._handler_funcs)

    def add_default_constraints(self):
        """Adds default constraints for this type. Override for your own types."""

        # Nothing here for default subscribers :)

    def check_constraints(self, action: Action) -> List[Constraint.Violation]:
        """Checks all constraints. All violations are returned."""
        res = []
        for con in self._constraints:
            r = con.check(action)
            if r is not None:
                res.append(r)
        return res


_HandlerFunc = Callable[[Subscriber, Event], Optional[List[Action]]]


def _assert_legal_handler(func: _HandlerFunc):
    """Checks whether the function can be used as an event handler.

    Notes
    -----
    The logic for this is very complicated. Here are the basic ideas.

    - The bare signature (without annotations) should be `func(self, event)`
    - The names must be "self" and "event"!
    - Allowed annotations for "self":
        - `self` (empty)
        - `self: ST` where issubclass(ST, Subscriber)
    - Allowed annotations for "event":
        - `event` (empty)
        - `event: ET` where issubclass(ET, Event)
    - Allowed annotations for the return type:
        - `): ` (empty)
        - `: -> None` (no returned value)
        - `: -> List[AT]` where issubclass(AT, Action)
        - `: -> Optional[List[AT]]` where issubclass(AT, Action)
        - `: -> Union[List[AT1], List[AT2]]` where issubclass(AT#, Action)

    """

    # This should always work
    sig = inspect.signature(func)
    try:
        p_self, p_event = list(sig.parameters.values())
        assert p_self.name == "self"
        assert p_event.name == "event"
    except Exception as e:
        raise MafiaBadHandler(func) from e

    # This may not work, if import schenanigans occur
    try:
        type_hints = get_type_hints(func)
    except Exception:
        logger.exception("Could not get type hints for event handler:")
        return

    # Assuming typing worked, we can check these
    try:
        # Argument "self"
        # `self`, `self: SubscriberSubtype`
        th_self = type_hints.get("self")
        if th_self is not None:
            assert issubclass(th_self, Subscriber)

        # Argument "event"
        # `event`, `event: EventSubtype`, `event: Union[Event1, Event2]`
        th_event = type_hints.get("event")
        if th_event is None:
            # No type hints
            pass
        elif get_origin(th_event) is None:
            # `event: Event`
            assert issubclass(th_event, Event)
        elif get_origin(th_event) is Union:
            # `event: Union[Event1, Event2]`
            for arg in get_args(th_event):
                assert issubclass(arg, Event)
        else:
            raise TypeError(f"Bad typing hint for 'event': {th_event!r}")

        # Return Type
        # `: `, `-> None`, `: -> List[Action]`, `: -> Optional[List[Action]]`
        th_return = type_hints.get("return")
        if th_return in [None, NoneType]:
            # None or no return type
            pass
        elif get_origin(th_return) is None:
            # This might just be `Action`, for example
            raise TypeError(f"Must return None or List[Action], got {th_return!r}")
        elif get_origin(th_return) == list:
            # List[Action]
            for arg in get_args(th_return):
                assert issubclass(arg, Action)
        elif get_origin(th_return) == Union:
            # Optional[List[Action]] or Union[List[Action1], List[Action2]]
            for arg in get_args(th_return):
                if arg in [None, NoneType]:
                    pass
                elif get_origin(arg) == list:
                    for a2 in get_args(arg):
                        assert issubclass(a2, Action)
                else:
                    raise TypeError(f"Union contains unsupported arg: {arg!r}")
    except Exception as e:
        raise MafiaBadHandler(func) from e


class EventHandler(object):
    """Descriptor that implements event handling logic."""

    def __init__(self, func: _HandlerFunc, *etypes: List[Event]):
        _assert_legal_handler(func)

        @wraps(func)
        def wrapped(self: Subscriber, event: Event) -> Optional[List[Action]]:
            # Wrap the function to check constraints.
            raw = func(self, event)
            if raw is None:
                return None
            # TODO: Maybe allow just returning Action or Optional[Action]?
            # That would simplify many handler functions that return just 1 Action.
            # But I'd have to rewrite _assert_legal_handler() ... :P
            if not isinstance(raw, list):
                raise TypeError(f"Expected List[Action], got {raw!r}")
            res = []
            for action in raw:
                violations = self.check_constraints(action)
                if len(violations) == 0:
                    res.append(action)
                else:
                    msg = [f"Constraint Violations for {type(action).__qualname__}:"]
                    msg += [f"  {v.msg}" for v in violations]
                    logger.warning("\n".join(msg))
            return res

        self.func: _HandlerFunc = wrapped  # func
        self.etypes: List[Event] = list(etypes)

    def __set_name__(self, owner: Type[Subscriber], name: str):
        if not issubclass(owner, Subscriber):
            raise TypeError(f"Expected Subscriber, got {owner!r}")

        self.name = name

    def __get__(self, obj: Subscriber, objtype=None) -> Callable:
        # This returns the method that was wrapped

        return partial(self.func, obj)
        # return partial(wrapped, obj)


def handler(func: _HandlerFunc) -> EventHandler:
    """Decorator to automatically infer event handler.

    Usage
    -----
    Use this as a decorator with a mandatory type hint for `event`:

        class A(Subscriber):
            @handler
            def f(self, event: Union[EPreAction, EPostAction]):
                return []
    """

    type_hints = get_type_hints(func)
    th = type_hints.get("event")
    if th is None:
        raise TypeError("Type hint for 'event' is required; otherwise, use `handles()`")
    if get_origin(th) is None:
        if issubclass(th, Event):
            return EventHandler(func, th)
    elif get_origin(th) is Union:
        etypes = get_args(th)
        for a in etypes:
            if not issubclass(a, Event):
                raise TypeError(f"One of Union types is not Event: {th!r}")
        return EventHandler(func, *etypes)
    raise NotImplementedError(f"Unsupported type hint: {th!r}")


def handles(*etypes: List[Event]) -> Callable[[_HandlerFunc], EventHandler]:
    """Decorator factory, to handle events.

    Usage
    -----
    Use this as a decorator factory with the event types as arguments:

        class A(Subscriber):
            @handles(EPreAction, EPostAction)
            def f(self, event) -> Optional[List[Action]]:
                return None
    """

    def _inner(func: _HandlerFunc) -> EventHandler:
        return EventHandler(func, *etypes)

    return _inner


class Constraint(Subscriber):
    """Base class for constraints.

    Override `check` and return a `self.Violation("Description")` or `None`.

    Note that each Constraint is itself a Subscriber, so it technically can have
    its own constraints, but by default they're not used.
    """

    def __init__(self, game, /, parent: Subscriber):
        self._parent = parent
        super().__init__(game, use_default_constraints=False)
        self._parent._constraints.append(self)  # FIXME: How do we remove constraints?

    @property
    def parent(self) -> Subscriber:
        return self._parent

    class Violation(object):
        """Constraint was violated."""

        def __init__(self, msg: str) -> None:
            self._msg = str(msg)

        @property
        def msg(self) -> str:
            return self._msg

        def __repr__(self) -> str:
            cn = type(self).__qualname__  # e.g. "Constraint.Violation"
            return f"{cn}({self.msg!r})"

    @abstractmethod
    def check(self, action: Action) -> Optional[Constraint.Violation]:
        """Checks whether the `action` is allowed.

        If allowed, return None.
        If not allowed, return a `Constraint.Violation`
        """

    def hook_pre_action(self, action: Action) -> Optional[List[Action]]:
        """Hook called when parent is trying to action & no violation for self."""

    def hook_post_action(self, action: Action) -> Optional[List[Action]]:
        """Hook called when parent successfully actioned."""

    @handler
    def handler_pre(self, event: EPreAction) -> Optional[List[Action]]:
        if isinstance(event, EPreAction) and event.action.source == self.parent:
            violation = self.check(event.action)
            if violation is None:
                return self.hook_pre_action(event.action)
            # we have a violation - cancel!
            return [CancelAction(self.game, self, target=event.action)]

    @handler
    def handler_post(self, event: EPostAction) -> Optional[List[Action]]:
        if isinstance(event, EPostAction) and event.action.source == self.parent:
            return self.hook_post_action(event.action)

    @property
    def prefix_tags(self) -> List[str]:
        """These are tags used for descriptions, as a prefix."""
        return []


class EventEngine(GameObject):
    """Subscription and broadcasting engine."""

    def __init__(self, game: Game):
        self._handlers: DefaultDict[Type[Event], List[Callable]] = defaultdict(list)
        self._subscribers: DefaultDict[Type[Event], List[Subscriber]] = defaultdict(
            list
        )
        super().__init__(game)

    def add_handler(self, handler: EventHandler, parent: Subscriber) -> _HandlerFunc:
        """Adds the handler, with given parent, to own subscribers."""
        f = partial(handler.func, parent)
        for etype in handler.etypes:
            self._handlers[etype].append(f)

        if parent not in self._handlers[etype]:
            self._subscribers[etype].append(parent)
        return f

    def remove_subscriber(self, sub: Subscriber):
        """Removes all subscriptions from the subscriber.

        FIXME: This operation is very hackish and iterates over EVERYTHING.
        This can probably be fixed by adding back-references, somehow.
        """
        hfs = sub.handler_funcs
        for etype in self._subscribers.keys():
            try:
                self._subscribers[etype].remove(sub)
                for hf in hfs:
                    try:
                        self._handlers[etype].remove(hf)
                    except ValueError:
                        pass
            except ValueError:
                pass
        sub._handler_funcs = []

    def broadcast(self, event: Event) -> List[Action]:
        """Broadcasts event to all handlers."""

        # Loop over superclasses, but make sure you don't repeat handlers
        funcs = []  # NOTE: not using a set, because we want deterministic sorting
        ET = type(event)
        for T in ET.mro():
            if issubclass(T, Event):
                funcs += [h for h in self._handlers[T] if h not in funcs]

        # Call each of the functions
        res = []
        for f in funcs:
            x = f(event)
            if x is None:
                x = []
            res.extend(x)
        return res



from typing import List

PATH_SEP = "/"  # Separates paths in hierarchical structures
# TODO: Consider other separators.

ABILITY = "ability"
TRIGGER = "trigger"


def get_parts(path: str) -> List[str]:
    if not isinstance(path, str):
        raise TypeError(f"Expected str, got {path!r}")
    return path.split(PATH_SEP)


def get_path(*parts: List[str]) -> str:
    return PATH_SEP.join(parts)



from __future__ import annotations

import logging
import shlex
from abc import ABC, abstractmethod
from typing import List

from .raw import RawCommand

logger = logging.getLogger(__name__)

__all__ = ["AbstractCommandParser", "ShellCommandParser"]


class AbstractCommandParser(ABC):
    """Base for command parsers."""

    @abstractmethod
    def parse(self, source: str, obj: str) -> List[RawCommand]:
        """Parses `obj` into zero, one or more `RawCommand` objects."""
        return []


class ShellCommandParser(AbstractCommandParser):
    """Implementation for shell-like command parsing."""

    def parse(self, source: str, obj: str) -> List[RawCommand]:
        if not isinstance(obj, str):
            # raise TypeError(f"Expected str, got {obj!r}")
            logger.warning(f"Cannot parse object, ignoring: {obj!r}")
            return []

        text = obj
        # TODO: keyword and flag arguments? Not just positional :)
        raw = shlex.split(text, posix=True)
        res = RawCommand(source, raw[0], args=tuple(raw[1:]))
        return [res]


if __name__ == "__main__":
    # For testing
    sp = ShellCommandParser()


"""Optional Pygments lexer for the Mafia 'shell'-style language."""

from pygments.lexer import RegexLexer
from pygments.token import *

__all__ = ["MafiaCliLexer", "ImplicitMafiaCliLexer"]


class MafiaCliLexer(RegexLexer):
    """Custom Pygments lexer for 'multi-user' Mafia command-line parser.

    Current syntax:

        USER COMMAND <arg> <arg>

    No argument validation is done, only quoting.
    """

    name = "MultiMafia"
    aliases = ["multi-mafia"]
    filenames = []

    tokens = {
        "root": [
            (R"^\s*", Text, "user"),
        ],
        "user": [
            (R'"', Name.Variable, "user-quote"),
            (R"\b\S+\b", Name.Variable, "command"),
        ],
        "user-quote": [
            ('[^"]+', Name.Variable),
            ('"', Name.Variable, "command"),
        ],
        "command": [
            (R'"', Name.Exception, "command-quote"),
            (R"\b\S+\b", Name.Exception, "args"),
        ],
        "command-quote": [
            ('[^"]+', Name.Exception),
            ('"', Name.Exception, "args"),
        ],
        "args": [
            (R'"', String.Double, "args-quote"),
            (R"\b.*\n", Text, "root"),
            (R"\s+\n", Text, "root"),
        ],
        "args-quote": [
            ('[^"]+', String.Double),
            ('"', String.Double, "#pop"),
        ],
    }


class ImplicitMafiaCliLexer(RegexLexer):
    """Custom Pygments lexer for 'single-user' Mafia command-line parser.

    Current syntax:

        COMMAND <arg> <arg>

    No argument validation is done, only quoting.
    """

    name = "ImplicitMafia"
    aliases = ["implicit-mafia", "single-mafia"]
    filenames = []

    tokens = {
        "root": [
            (R"^\s*", Text, "command"),
        ],
        "command": [
            (R'"', Name.Exception, "command-quote"),
            (R"\b\S+\b", Name.Exception, "args"),
        ],
        "command-quote": [
            ('[^"]+', Name.Exception),
            ('"', Name.Exception, "args"),
        ],
        "args": [
            (R'"', String.Double, "args-quote"),
            (R"\b.*\n", Text, "root"),
            (R"\s+\n", Text, "root"),
        ],
        "args-quote": [
            ('[^"]+', String.Double),
            ('"', String.Double, "#pop"),
        ],
    }



import warnings

from open_mafia_engine.core.game import Game
from open_mafia_engine.core.game_object import converter
from open_mafia_engine.core.naming import ABILITY, PATH_SEP, TRIGGER, get_parts
from open_mafia_engine.core.phase_cycle import Phase
from open_mafia_engine.core.state import Ability, Actor, Faction, Trigger
from open_mafia_engine.util.matcher import FuzzyMatcher


@converter.register
def get_faction_by_name(game: Game, obj: str) -> Faction:
    """Gets the Faction by exact or fuzzy name match."""

    matcher = FuzzyMatcher({f.name: f for f in game.factions}, score_cutoff=20)
    try:
        return matcher[obj]
    except Exception as e:
        raise ValueError(f"Could not find Faction by name: {obj!r}") from e


@converter.register
def get_actor_by_name(game: Game, obj: str) -> Actor:
    """Gets the Actor by exact or fuzzy name match."""
    matcher = FuzzyMatcher({a.name: a for a in game.actors}, score_cutoff=10)
    try:
        return matcher[obj]
    except Exception as e:
        raise ValueError(f"Could not find Actor by name: {obj!r}") from e


@converter.register
def get_ability_by_path(game: Game, obj: str) -> Ability:
    """Gets the ability by 'path' made of names.

    Assuming PATH_SEP is "/", this will parse as:

        "{actor_name}/ability/{ability_name}"

    This will do fuzzy matching on Actor and Ability separately.
    TODO: Maybe do fuzzy matching on total string?
    """
    try:
        owner_name, _abil, abil_name = get_parts(obj)

        assert _abil == ABILITY
    except Exception as e:
        raise ValueError(f"Bad/non-existing path for Ability: {obj}") from e

    owner: Actor = get_actor_by_name(game, owner_name)
    matcher = FuzzyMatcher({ab.name: ab for ab in owner.abilities}, score_cutoff=10)
    try:
        return matcher[abil_name]
    except Exception as e:
        raise ValueError(f"Could not find Ability by path: {obj!r}") from e


@converter.register
def get_trigger_by_path(game: Game, obj: str) -> Trigger:
    """Gets the trigger by 'path' made of names.

    Assuming PATH_SEP is "/", this will parse as:

        "{actor_name}/trigger/{trigger_name}"

    This will do fuzzy matching on Actor and Trigger separately.
    TODO: Maybe do fuzzy matching on total string?
    """
    try:
        owner_name, _trig, trig_name = get_parts(obj)

        assert _trig == TRIGGER
    except Exception as e:
        raise ValueError(f"Bad/non-existing path for Trigger: {obj}") from e

    owner: Actor = get_actor_by_name(game, owner_name)
    matcher = FuzzyMatcher({tr.name: tr for tr in owner.triggers}, score_cutoff=10)
    try:
        return matcher[trig_name]
    except Exception as e:
        raise ValueError(f"Could not find Trigger by path: {obj!r}") from e


@converter.register
def get_phase_by_name(game: Game, obj: str) -> Phase:
    """Gets the phase by name from the cycle. Can raise KeyError."""
    try:
        return game.phase_system[obj]
    except KeyError:
        pass

    # Make sure we don't infinitely loop through phases
    N_MAX = 20
    options = {}
    for i, p in enumerate(game.phase_system.possible_phases):
        if i > N_MAX:
            warnings.warn(f"Found over {N_MAX} phases; maybe infinite?")
            break
        options[p.name] = p
    matcher = FuzzyMatcher(options, score_cutoff=50)
    return matcher[obj]



from open_mafia_engine.core.all import ATBase, Ability, Action, Actor, Game, GameObject


class BaseInformationAction(Action):
    """Action that gives information to a player.

    Only ATBase (e.g. Ability or Trigger) may create this action,
    so we always know the "owner" (who to inform).

    Inherit from this action.
    """

    def __init__(
        self, game: Game, source: ATBase, /, *, priority: float, canceled: bool
    ):
        if not isinstance(source, ATBase):
            raise TypeError(f"Information actions require players.")
        super().__init__(game, source, priority=priority, canceled=canceled)

    @property
    def source(self) -> ATBase:
        return self._source

    @property
    def owner(self) -> Actor:
        return self.source.owner

    @classmethod
    def status_key(self) -> str:
        """Returns the key where results will be stored."""
        return "info_results"

    def inform(self, msg: str):
        """Helper function to inform the user.

        This will trigger `EStatusChanged`.
        """
        msg = str(msg)

        key = self.status_key()
        st = self.owner.status
        old = st.get(key)
        if old is None:
            st[key] = [msg]
        elif isinstance(old, list):
            st[key] = old + [msg]
        else:
            raise TypeError(f"Expected previous status to be list, but got: {old!r}")

    # doit()


class BaseInspectAction(BaseInformationAction):
    """Action that inspects an Actor.

    Inherit from this action.
    """

    def __init__(
        self,
        game: Game,
        source: ATBase,
        /,
        target: Actor,
        *,
        priority: float = 0.0,
        canceled: bool = False,
    ):
        self._target = target
        super().__init__(game, source, priority=priority, canceled=canceled)

    @property
    def target(self) -> Actor:
        return self._target

    @target.setter
    def target(self, v: Actor):
        self._target = v

    # doit()


class InspectFactionAction(BaseInspectAction):
    """Inspect the factions of a player."""

    def doit(self):
        found = [f.name for f in self.target.factions]
        msg = "Your target is part of these factions: " + ", ".join(found)
        self.inform(msg)


InspectFactionAbility = Ability.generate(
    InspectFactionAction,
    params=["target"],
    name="InspectFactionAbility",
    doc="Ability to inspect someone's faction(s).",
    desc="Inspect target's faction(s).",
)



from __future__ import annotations

from typing import TYPE_CHECKING

from open_mafia_engine.core.enums import Outcome
from open_mafia_engine.core.event_system import Action, EPostAction, EPreAction
from open_mafia_engine.core.game_object import GameObject

if TYPE_CHECKING:
    from open_mafia_engine.core.game import Game
    from open_mafia_engine.core.state import Faction


class EOutcomeAchieved(EPostAction):
    """An outcome has been achieved."""

    def __init__(self, game, action: OutcomeAction, /):
        super().__init__(game, action)

    @property
    def action(self) -> OutcomeAction:
        return self._action

    @property
    def faction(self) -> Faction:
        return self.action.faction

    @property
    def outcome(self) -> Outcome:
        return self.action.outcome


class OutcomeAction(Action):
    """A Faction achieves victory or defeat."""

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        faction: Faction,
        outcome: str,
        *,
        priority: float = 100.0,
        canceled: bool = False,
    ):
        super().__init__(game, source, priority=priority, canceled=canceled)
        self._faction = faction
        self._outcome = Outcome(outcome)

    def doit(self):
        """Sets the "outcome" status for each Actor in Faction?"""
        for actor in self.faction.actors:
            actor.status["outcome"] = self.outcome

    class Pre(EPreAction):
        """A Faction is about to achieve an Outcome."""

    Post = EOutcomeAchieved

    @property
    def faction(self) -> Faction:
        return self._faction

    @faction.setter
    def faction(self, v):
        from open_mafia_engine.core.state import Faction

        if not isinstance(v, Faction):
            raise TypeError(f"Expected Faction, got {v!r}")
        self._faction = v

    @property
    def outcome(self) -> Outcome:
        return self._outcome

    @outcome.setter
    def outcome(self, v: str):
        self._outcome = Outcome(v)



from open_mafia_engine.builders.for_testing import make_test_game
from open_mafia_engine.core.game import Game
from open_mafia_engine.core.game_object import inject_converters
from open_mafia_engine.core.state import Actor


def test_fuzzy_actor_converter():
    """Tests fuzzy matching for Actor converter."""

    game = make_test_game(["Alice", "Bob"])

    @inject_converters
    def hey(game: Game, actor: Actor) -> str:
        return f"Hello, {actor.name}"

    assert hey(game, "alice") == "Hello, Alice"



from typing import List, Optional
import warnings
from open_mafia_engine.core.all import (
    Ability,
    Action,
    ActionInspector,
    Actor,
    ATBase,
    EPreAction,
    Game,
    GameObject,
    handler,
)
from open_mafia_engine.util.matcher import FuzzyMatcher

from .auxiliary import TempPhaseAux


class BaseRedirectAction(Action):
    """Action that redirects another action to a new target.

    Note that actions can have multiple targets.
    This redirects the one closest to `field_name`
    """

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        target: Action,
        new_target: GameObject,
        field_name: str = "target",
        *,
        priority: float = 50.0,
        canceled: bool = False,
    ):
        self._target = target
        self._field_name = str(field_name)
        self._new_target = new_target
        super().__init__(game, source, priority=priority, canceled=canceled)

    @property
    def target(self) -> Action:
        return self._target

    @target.setter
    def target(self, v: Action):
        if not isinstance(v, Action):
            raise TypeError(f"Can only redirect Actions, but got {v!r}")
        self._target = v

    @property
    def new_target(self) -> GameObject:
        return self._new_target

    # setter?

    @property
    def field_name(self) -> str:
        return self._field_name

    def doit(self):
        action = self.target
        ai = ActionInspector(action)

        # Look for name match, with high confidence
        p1 = {p: p for p in ai.param_names}
        found = FuzzyMatcher(p1, score_cutoff=80).get(self.field_name, None)

        if found is None:
            # If not, check for exact type matches
            type_matches = ai.params_of_type(type(self.new_target))
            p2 = {p: p for p in type_matches}
            found = FuzzyMatcher(p2, score_cutoff=0).get(self.field_name, None)

        if found is None:
            # Meh - we failed, lets set it anyways for history.
            warnings.warn(
                f"Could not redirect. Field name: {self.field_name!r}."
                f" Action: {action!r}. New target: {self.new_target!r}."
            )
            found = self.field_name

        ai.set_value(found, self.new_target)


class ActorRedirectAction(BaseRedirectAction):
    """Redirects actions to"""

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        target: Action,
        new_target: Actor,
        field_name: str = "target",
        *,
        priority: float = 50.0,
        canceled: bool = False,
    ):
        super().__init__(
            game,
            source,
            target=target,
            new_target=new_target,
            field_name=field_name,
            priority=priority,
            canceled=canceled,
        )

    @property
    def new_target(self) -> Actor:
        return self._new_target

    # setter?


class ActorRedirectorAux(TempPhaseAux):
    """Aux object that redirects actions made by the target to another actor.

    It removes itself after the end of the phase.

    Attributes
    ----------
    game : Game
    target : Actor
        The target to redirect.
    new_target : Actor
        The resulting target.
    field_name : str
        The field name to search for.
    key : None or str
        Safely ignore this (?).
    only_abilities : bool
        If True, only blocks Ability activations, and lets Triggers through.
        Default is True.
    """

    def __init__(
        self,
        game: Game,
        /,
        target: Actor,
        new_target: Actor,
        field_name: str = "target",
        key: str = None,
        *,
        only_abilities: bool = True,
        use_default_constraints: bool = True,
    ):
        self._only_abilities = bool(only_abilities)
        self._target = target
        self._new_target = new_target
        self._field_name = str(field_name)
        super().__init__(
            game, key=None, use_default_constraints=use_default_constraints
        )

    @property
    def target(self) -> Actor:
        return self._target

    @property
    def new_target(self) -> Actor:
        return self._new_target

    @property
    def field_name(self) -> str:
        return self._field_name

    @property
    def only_abilities(self) -> bool:
        return self._only_abilities

    @handler
    def handle_to_cancel(
        self, event: EPreAction
    ) -> Optional[List[ActorRedirectAction]]:
        """Cancels the action if it came from the target."""
        if isinstance(event, EPreAction):
            src = event.action.source
            if isinstance(src, ATBase):
                if self.only_abilities and not isinstance(src, Ability):
                    # Skip
                    return
                if src.owner == self.target:
                    return [
                        ActorRedirectAction(
                            self.game,
                            self,
                            target=event.action,
                            new_target=self.new_target,
                            field_name=self.field_name,
                        )
                    ]


class CreateRedirectAction(Action):
    """Action that redirects all actions from the target to new_target."""

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        target: Actor,
        new_target: Actor,
        field_name: str = "target",
        *,
        priority: float = 90,
        canceled: bool = False,
    ):
        self.target = target
        self.new_target = new_target
        self._field_name = str(field_name)
        super().__init__(game, source, priority=priority, canceled=canceled)

    @property
    def field_name(self) -> str:
        return self._field_name

    def doit(self):
        ActorRedirectorAux(
            self.game,
            target=self.target,
            new_target=self.new_target,
            field_name=self.field_name,
            only_abilities=True,
        )


CreateRedirectAbility = Ability.generate(
    CreateRedirectAction,
    params=["target", "new_target"],
    name="CreateRedirectAbility",
    doc="Ability to redirect others.",
    desc="Redirects the target to another target.",
)


from typing import List, Optional

from open_mafia_engine.core.all import (
    Ability,
    Action,
    Actor,
    ATBase,
    CancelAction,
    EPreAction,
    Game,
    GameObject,
    handler,
)

from .auxiliary import TempPhaseAux


class RoleBlockerAux(TempPhaseAux):
    """Aux object that blocks actions made by the target.

    It removes itself after the end of the phase.

    Attributes
    ----------
    game : Game
    target : Actor
        The target to block.
    key : None or str
        Safely ignore this (?).
    only_abilities : bool
        If True, only blocks Ability activations, and lets Triggers through.
        Default is True.
    """

    def __init__(
        self,
        game: Game,
        /,
        target: Actor,
        key: str = None,
        *,
        only_abilities: bool = True,
        use_default_constraints: bool = True,
    ):
        self._only_abilities = bool(only_abilities)
        self._target = target
        super().__init__(
            game, key=None, use_default_constraints=use_default_constraints
        )

    @property
    def target(self) -> Actor:
        return self._target

    @property
    def only_abilities(self) -> bool:
        return self._only_abilities

    @handler
    def handle_to_cancel(self, event: EPreAction) -> Optional[List[CancelAction]]:
        """Cancels the action if it came from the target."""
        if isinstance(event, EPreAction):
            src = event.action.source
            if isinstance(src, ATBase):
                if self.only_abilities and not isinstance(src, Ability):
                    # Skip
                    return
                if src.owner == self.target:
                    return [CancelAction(self.game, self, target=event.action)]


class RoleBlockAction(Action):
    """Action that prevents the target from actioning until the end of the phase."""

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        target: Actor,
        *,
        priority: float = 90,
        canceled: bool = False,
    ):
        self.target = target
        super().__init__(game, source, priority=priority, canceled=canceled)

    def doit(self):
        RoleBlockerAux(self.game, target=self.target, only_abilities=True)


RoleBlockAbility = Ability.generate(
    RoleBlockAction,
    params=["target"],
    name="RoleBlockAbility",
    doc="Ability to block others.",
    desc="Roleblocks the target.",
)



from __future__ import annotations

import logging
from abc import abstractmethod
from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional, Tuple, Union

from open_mafia_engine.core.all import (
    Ability,
    Action,
    Actor,
    AuxObject,
    EPostAction,
    EPostPhaseChange,
    EPreAction,
    Game,
    GameObject,
    PhaseChangeAction,
    converter,
    get_actor_by_name,
    handler,
    inject_converters,
)

NoneType = type(None)
logger = logging.getLogger(__name__)

# TODO: tiebreaking enum, hammers, allow unvotes, allow against all


class AbstractVoteTarget(GameObject):
    """A target for voting. Don't use directly."""

    @abstractmethod
    def apply(self, vr: VotingResults, voter: Actor):
        """Apply self on the voting results."""

    def __eq__(self, o: object) -> bool:
        # NOTE: This can be done better :)
        if not isinstance(o, AbstractVoteTarget):
            return NotImplemented
        if self.game != o.game:
            return False
        return repr(self) == repr(o)

    def __hash__(self) -> int:
        return hash(repr(self))


class VotingOptions(GameObject):
    """Voting options.

    Attributes
    ----------
    game : Game
    allow_unvote : bool
        Whether unvotes are allowed at all. Default is True.
    allow_against_all : bool
        Whether a "VoteAgainstAll" (e.g. "no lynch") is allowed. Default is True.

    TODO: Hammers? Tiebreaks?
    """

    def __init__(
        self,
        game: Game,
        /,
        *,
        allow_unvote: bool = True,
        allow_against_all: bool = True,
    ):
        super().__init__(game)
        self._allow_unvote = bool(allow_unvote)
        self._allow_against_all = bool(allow_against_all)

    @property
    def allow_unvote(self) -> bool:
        return self._allow_unvote

    @property
    def allow_against_all(self) -> bool:
        return self._allow_against_all


class VotingResults(GameObject):
    """Voting results object."""

    def __init__(
        self,
        game: Game,
        options: VotingOptions,
        /,
    ):
        super().__init__(game)
        self._voters: List[Actor] = []
        self._targets: List[GameObject] = []
        self._options = options

        def mk():
            return defaultdict(float)

        # voter -> target -> qty
        self._map: DefaultDict[int, DefaultDict[int, float]] = defaultdict(mk)

    @property
    def options(self) -> VotingOptions:
        return self._options

    @property
    def vote_map(self) -> List[Tuple[GameObject, float, List[Actor]]]:
        """Gets the vote counts, along with Actors who vote for them."""
        res = []
        for i_t, t in enumerate(self._targets):
            voters = []
            total = 0
            for k, v in self._map.items():
                qty = v[i_t]
                total += qty
                if qty > 0:
                    voters.append(self._voters[k])
            res.append((t, total, voters))
        res = sorted(res, key=lambda x: -x[1])
        return res

    @property
    def vote_counts(self) -> List[Tuple[GameObject, float]]:
        """Gets the vote counts, sorted in descending order."""
        res = []
        for i_t, t in enumerate(self._targets):
            total = 0
            for _, v in self._map.items():
                total += v[i_t]
            res.append((t, total))
        res = sorted(res, key=lambda x: -x[1])
        return res

    @property
    def vote_leaders(self) -> List[GameObject]:
        """Gets the current vote leaders, if any."""
        vc = self.vote_counts
        if len(vc) == 0:
            return []

        max_cnt = vc[0][1]
        # This short-circuits any votes :)
        if max_cnt <= 0:
            return []

        res = []
        for target, cnt in vc:
            if cnt < max_cnt:
                break
            res.append(target)
        return res

    @inject_converters
    def _i_voter(self, voter: Actor) -> int:
        if voter in self._voters:
            i_voter = self._voters.index(voter)
        else:
            i_voter = len(self._voters)
            self._voters.append(voter)
        return i_voter

    @inject_converters
    def _i_target(self, target: GameObject) -> int:
        if target in self._targets:
            i_target = self._targets.index(target)
        else:
            i_target = len(self._targets)
            self._targets.append(target)
        return i_target

    def _reset(self, voter: Actor):
        i_src = self._i_voter(voter)
        self._map[i_src] = defaultdict(float)

    def _set_i(self, i_voter: int, i_target: int, weight: float = 1):
        self._map[i_voter][i_target] = weight

    def _get_i(self, i_voter: int, i_target: int) -> float:
        return self._map[i_voter][i_target]

    def _set(self, voter: Actor, target: GameObject, weight: float = 1):
        self._set_i(self._i_voter(voter), self._i_target(target), weight=weight)

    def _get(self, voter: Actor, target: GameObject) -> float:
        return self._get_i(self._i_voter(voter), self._i_target(target))


class UnvoteAll(AbstractVoteTarget):
    """Symbolic class for unvoting."""

    def apply(self, vr: VotingResults, voter: Actor):
        """Removes all votes made by `voter`."""
        if not vr.options.allow_unvote:
            # Warn?
            return
        vr._reset(voter)

    def __eq__(self, o: object) -> bool:
        if isinstance(o, UnvoteAll) and self.game == o.game:
            return True
        return False


class VoteAgainstAll(AbstractVoteTarget):
    """Symbolic class for voting against all options."""

    def apply(self, vr: VotingResults, voter: Actor):
        """Unvotes-all, then votes for `VoteAgainstAll` object."""
        if not vr.options.allow_against_all:
            # Warn?
            return
        vr._reset(voter)
        vr._set(
            voter,
            VoteAgainstAll(self.game),
        )

    def __eq__(self, o: object) -> bool:
        if isinstance(o, VoteAgainstAll) and self.game == o.game:
            return True
        return False


class ActorTarget(AbstractVoteTarget):
    """Normal voting for a single Actor."""

    def __init__(self, game, /, actor: Actor):
        super().__init__(game)
        self._actor = actor

    @property
    def actor(self) -> Actor:
        return self._actor

    def apply(self, vr: VotingResults, voter: Actor):
        vr._reset(voter)
        vr._set(voter, self.actor)


class ActorTargets(AbstractVoteTarget):
    """Voting for multiple actors."""

    def __init__(self, game, /, actors: List[Actor]):
        super().__init__(game)
        self._actors = list(actors)

    @property
    def actors(self) -> List[Actor]:
        return list(self._actors)

    def apply(self, vr: VotingResults, voter: Actor):
        vr._reset(voter)
        for a in self.actors:
            vr._set(voter, a)


# TODO: Weighted votes? Weighted targets?


class Vote(GameObject):
    """A vote on a Tally.

    Attributes
    ----------
    voter : Actor
        The actor who is voting.
    target : AbstractVoteTarget
        The target, or targets, or UnvoteAll...
    """

    def __init__(self, game: Game, /, voter: Actor, target: AbstractVoteTarget):
        super().__init__(game)
        self.voter = voter
        self.target = target

    @property
    def voter(self) -> Actor:
        return self._source

    @voter.setter
    @inject_converters
    def voter(self, v: Actor):
        if not isinstance(v, Actor):
            raise TypeError(f"Expected Actor, got {v!r}")
        self._source = v

    @property
    def target(self) -> AbstractVoteTarget:
        return self._target

    @target.setter
    @inject_converters
    def target(self, v: AbstractVoteTarget):
        if not isinstance(v, AbstractVoteTarget):
            raise TypeError(f"Expected AbstractVoteTarget, got {v!r}")
        self._target = v


class VoteAction(Action):
    """Votes for someone."""

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        voter: Actor,
        target: AbstractVoteTarget,
        tally: Tally = None,
        *,
        priority: float = 0.0,
        canceled: bool = False,
    ):
        super().__init__(game, source, priority=priority, canceled=canceled)
        self.voter = voter
        self.target = target
        if tally is None:
            tally = get_default_tally(game, tally)
        self.tally = tally

    def doit(self):
        self.tally.add_vote(Vote(self.game, voter=self.voter, target=self.target))

    class Pre(EPreAction):
        """We are about to vote."""

    class Post(EPostAction):
        """We have voted."""


class Tally(AuxObject):
    """Voting tally.

    Tallies select the leader before each phase end. Override `respond_leader()`.
    Tallies reset after each phase end.
    """

    def __init__(
        self,
        game: Game,
        /,
        *,
        allow_unvote: bool = True,
        allow_against_all: bool = True,
    ):
        super().__init__(game)
        self._vote_history: List[Vote] = []
        self._options = VotingOptions(
            game, allow_unvote=allow_unvote, allow_against_all=allow_against_all
        )

    @abstractmethod
    def respond_leader(self, leader: GameObject) -> Optional[List[Action]]:
        """Override this for particular behavior."""
        return None

    @property
    def options(self) -> VotingOptions:
        return self._options

    @property
    def allow_unvote(self) -> bool:
        return self.options.allow_unvote

    @property
    def allow_against_all(self) -> bool:
        return self.options.allow_against_all

    @property
    def vote_history(self) -> List[Vote]:
        return list(self._vote_history)

    @property
    def results(self) -> VotingResults:
        """Applies vote history to get current results."""
        res = VotingResults(self.game, self.options)
        for v in self.vote_history:
            v.target.apply(res, v.voter)
        return res

    def add_vote(self, vote: Vote):
        """Adds the vote to history."""
        if not isinstance(vote, Vote):
            raise TypeError(f"Can only add a Vote, got {vote!r}")
        self._vote_history.append(vote)

    @handler
    def reset_on_phase(self, event: EPostPhaseChange):
        """Resets votes every phase change."""
        self._vote_history = []

    @handler
    def handle_leader(self, event: PhaseChangeAction.Pre):
        leaders = self.results.vote_leaders
        if len(leaders) == 0:
            return
        elif len(leaders) > 1:
            return  # TODO: Tiebreaks, but they must be deterministic!
        leader = leaders[0]
        return self.respond_leader(leader)

    @handler
    def check_hammer(self, event: VoteAction.Post):
        """TODO: Hammers."""

        # Check if hammers are set.
        # If yes, check if this vote hammers.
        # If yes, end the phase.


class VoteAbility(Ability):
    """Voting ability."""

    def __init__(
        self,
        game: Game,
        /,
        owner: Actor,
        name: str,
        tally: Tally = None,
        desc: str = "Voting ability.",
    ):
        super().__init__(game, owner, name, desc=desc)
        self.tally = tally  # Is this found by the Injector?...

    def activate(self, target: AbstractVoteTarget) -> Optional[List[VoteAction]]:
        """Creates the action."""

        # TODO: Constraints!

        try:
            return [
                VoteAction(
                    self.game, self, voter=self.owner, target=target, tally=self.tally
                )
            ]
        except Exception:
            logger.exception("Error executing action:")
            if False:  # True, if debugging?
                raise
            return None


@converter.register
def get_default_tally(game: Game, obj: NoneType) -> Tally:
    """Selects the default vote tally."""
    tallies: List[Tally] = game.aux.filter_by_type(Tally)
    if len(tallies) == 0:
        raise TypeError(f"No tallies found!")
    elif len(tallies) > 1:
        raise TypeError(f"Multiple tallies found: {tallies!r}")
    return tallies[0]


@converter.register
def get_vote_target_actor(game: Game, obj: Actor) -> AbstractVoteTarget:
    return ActorTarget(game, actor=obj)


@converter.register
def get_vote_target_single(game: Game, obj: str) -> AbstractVoteTarget:
    """Parses a string as a vote target."""
    # TODO: Set phrases as game-specific options?
    if obj.lower() == "unvote":
        return UnvoteAll(game)
    elif obj.lower() == "no lynch":
        return VoteAgainstAll(game)
    return ActorTarget(game, actor=obj)  # will auto-convert the str, or fail.


@converter.register
def get_vote_target_multi(game: Game, obj: list) -> AbstractVoteTarget:
    """Parses a list of strings (or Actors) as a complex vote target."""
    # TODO: What if there are unvotes inside?...
    actors = [get_actor_by_name(game, x) for x in obj]
    return ActorTargets(game, actors=actors)


def get_vote_target(game: Game, obj: Union[str, list]) -> AbstractVoteTarget:
    if isinstance(obj, str):
        return get_vote_target_single(game, obj)
    elif isinstance(obj, list):
        return get_vote_target_multi(game, obj)
    raise TypeError(f"Can't get vote target for: {obj!r}")



from abc import abstractmethod
from typing import List

from open_mafia_engine.core.all import (
    Actor,
    EPrePhaseChange,
    Faction,
    Game,
    handler,
    Event,
)

from .auxiliary import TempPhaseAux


class ECreateFactionChat(Event):
    """Event that signals creating a faction chat."""

    def __init__(self, game: Game, /, faction: Faction):
        super().__init__(game)
        self.faction = faction

    @property
    def actors(self) -> List[Actor]:
        return self.faction.actors


class FactionChatCreatorAux(TempPhaseAux):
    """Base class to create the faction chat for some faction."""

    def __init__(self, game: Game, /, faction: Faction):
        self.faction = faction
        key = f"create chat for {self.faction.name}"
        super().__init__(game, key=key, use_default_constraints=False)

    @handler
    def handle_startup(self, event: EPrePhaseChange):
        if event.old_phase != self.game.phase_system.startup:
            return
        # NOTE: Rather than create an action, since it's startup, we should
        # just be able to trigger event responses.
        self.game.process_event(ECreateFactionChat(self.game, self.faction))

    @property
    def actors(self) -> List[Actor]:
        return self.faction.actors



"""Tests for building games."""

from open_mafia_engine.builders.for_testing import make_test_game
from open_mafia_engine.core.builder import GameBuilder


def test_builder_direct():
    """Tests the `test_builder()` function directly."""

    actor_names = ["Wendy", "Pickle"]
    game = make_test_game(actor_names)
    assert game.actor_names == actor_names


def test_builder_by_name():
    """Tests building the game."""
    from open_mafia_engine.builders.for_testing import make_test_game  # noqa

    builder = GameBuilder.load("test")
    game = builder.build(["Alice", "Bob"])
    assert game.actor_names == ["Alice", "Bob"]


def test_builder_args():
    """Tests building the game."""
    from open_mafia_engine.builders.for_testing import make_test_game  # noqa

    names = ["Alice", "Bob", "Charlie", "Dave", "Emily"]
    builder = GameBuilder.load("test")
    game = builder.build(names, n_mafia=2)
    assert game.actor_names == names



from enum import Enum
from typing import List, Type


def make_str_enum(name: str, values: List[str], doc: str = "") -> Type[str]:
    """Makes a string enumeration with particuluar values."""

    T = Enum(name, {v: v for v in values}, qualname=name, type=str)

    def __repr__(self: T) -> str:
        cn = type(self).__qualname__
        return f"{cn}({self.value!r})"

    T.__repr__ = __repr__
    T.__doc__ = doc

    return T



"""Tests for various longer scenarios."""

from open_mafia_engine.core.all import *
from open_mafia_engine.built_in.all import *
from open_mafia_engine.builders.all import *


def test_scenario_1():
    """Tests a scenario where lots of night roles are at play."""

    builder = GameBuilder.load("test")
    game = builder.build(["Alice", "Bob", "Charlie", "Dave", "Emily"], n_mafia=1)

    alice, bob, charlie, dave, emily = game.actors[:5]
    tally: Tally = game.aux.filter_by_type(Tally)[0]

    # Alice: Vote, Mafia Kill
    # Bob: Vote
    # Charlie: Vote, Protect
    # Dave: Vote, Inspect
    res_key = InspectFactionAction.status_key()

    # STARTUP
    # nothing yet
    game.change_phase()

    # DAY 1
    game.process_event(EActivate(game, "Alice/ability/Vote", target="Bob"))
    game.process_event(EActivate(game, "Bob/ability/Vote", target="Alice"))

    game.change_phase()
    assert not any(x.status["dead"] for x in game.actors)  # no-lynch, nobody dies

    # NIGHT 1
    # Mafia Alice tries to kill Bob
    game.process_event(EActivate(game, "Alice/ability/Mafia Kill", target=bob))
    # ... but Charlie protected bob, which has higher priority
    game.process_event(EActivate(game, "Charlie/ability/Protect", target=bob))
    # Dave inspects Alice
    game.process_event(EActivate(game, "Dave/ability/Faction Inspect", target=alice))
    # ... but is redirected by Emily to Charlie, and so inspects as Town!
    game.process_event(
        EActivate(game, "Emily/ability/Redirect", target=dave, new_target=charlie)
    )

    game.change_phase()  # process the night phase

    # nobody died
    assert not bob.status["dead"]
    assert not any(x.status["dead"] for x in game.actors)

    # Dave got 1 result, but they got redirected! They think Alice is Town.
    dave_n1_results = dave.status[res_key]
    assert isinstance(dave_n1_results, list)
    assert len(dave_n1_results) == 1
    assert "town" in dave_n1_results[0].lower()

    # DAY 2


def test_game_ending():
    """Simple test for the game ending properly."""

    game = make_test_game(["Alice", "Bob"])
    alice = game.actors[0]
    bob = game.actors[1]

    # Test whether alive constraint works :)
    alice.status["dead"] = True
    assert any(isinstance(a, EndTheGame) for a in game.action_queue.history)



from __future__ import annotations

from typing import Dict, List, Optional

from open_mafia_engine.core.auxiliary import AuxObject
from open_mafia_engine.core.converters import get_faction_by_name
from open_mafia_engine.core.enums import Outcome
from open_mafia_engine.core.event_system import Action, EPostAction, EPreAction, handler
from open_mafia_engine.core.game_object import GameObject, inject_converters
from open_mafia_engine.core.outcome import EOutcomeAchieved
from open_mafia_engine.core.state import Faction


class EGameEnded(EPostAction):
    """The game has ended."""

    @property
    def action(self) -> EndTheGame:
        return super().action

    @property
    def outcomes(self) -> Dict[str, Outcome]:
        return self.action.outcomes

    @inject_converters
    def outcome_for(self, faction: Faction) -> Outcome:
        """Returns the outcome for a given faction."""
        return self.outcomes[faction.name]

    @property
    def victorious_factions(self) -> List[Faction]:
        fns = [k for k, v in self.outcomes.items() if v == Outcome.victory]
        return [get_faction_by_name(self.game, fn) for fn in fns]

    @property
    def defeated_factions(self) -> List[Faction]:
        fns = [k for k, v in self.outcomes.items() if v == Outcome.defeat]
        return [get_faction_by_name(self.game, fn) for fn in fns]


class EndTheGame(Action):
    """Action that ends the game.

    Maybe this should inherit from PhaseChangeAction?
    I'm skeptical. Ending the game is significantly different.
    """

    def __init__(
        self,
        game,
        source: GameObject,
        /,
        outcomes: Dict[str, Outcome] = None,
        *,
        priority: float = 999,
        canceled: bool = False,
    ):
        if outcomes is None:
            outcomes = {}
        self.outcomes: Dict[str, Outcome] = dict(outcomes)
        super().__init__(game, source, priority=priority, canceled=canceled)

    class Pre(EPreAction):
        """The game is about to end."""

    Post = EGameEnded

    def doit(self):
        """Ends the game, setting the phase to 'shutdown'."""
        self.game.change_phase(self.game.phase_system.shutdown)


class GameEnder(AuxObject):
    """Ends the game when all factions get an Outcome."""

    def __init__(self, game, /):
        super().__init__(game)
        self._outcomes: Dict[str, Outcome] = {}

    @property
    def outcomes(self) -> Dict[str, Outcome]:
        return dict(self._outcomes)

    @handler
    def handle_outcome(self, event: EOutcomeAchieved) -> Optional[List[EndTheGame]]:
        """Checks off an outcome. If all factions have an outcome, ends the game."""
        if not isinstance(event, EOutcomeAchieved):
            return None
        self._outcomes[event.faction.name] = event.outcome

        if all(self.outcomes.get(fac) is not None for fac in self.game.faction_names):
            return [EndTheGame(self.game, self, self.outcomes)]



from typing import Dict, List, Optional
from uuid import uuid4

from open_mafia_engine.core.all import (
    Action,
    ActionInspector,
    ATBase,
    ATConstraint,
    AuxObject,
    Constraint,
    ConstraintActorTargetsAlive,
    ConstraintOwnerAlive,
    Game,
    Phase,
    PhaseChangeAction,
    Subscriber,
    handler,
)
from open_mafia_engine.core.event_system import EPostAction, EPreAction
from open_mafia_engine.core.state import Actor

from .auxiliary import CounterPerPhaseAux


class PhaseConstraint(Constraint):
    """Allow only using in a particular phase."""

    def __init__(self, game, /, parent: Subscriber, phase: Phase):
        self._phase = phase
        super().__init__(game, parent)

    @property
    def phase(self) -> Phase:
        return self._phase

    def check(self, action: Action) -> Optional[Constraint.Violation]:
        cp = self.game.current_phase
        if cp != self.phase:
            return self.Violation(
                f"Wrong phase: currently {cp.name!r}, need {self.phase.name!r}."
            )

    @property
    def prefix_tags(self) -> List[str]:
        """These are tags used for descriptions, as a prefix."""
        return [self.phase.name]


class LimitPerPhaseKeyConstraint(Constraint):
    """Allow only N uses per phase, given a key.

    Attributes
    ----------
    game : Game
    parent : Subscriber
    key : str
        They key to use for the aux object.
    limit : int
        The maximum number of actions per phase. Default is 1.
    only_successful : bool
        Whether to count only successful tries towards the limit.
        Default is False (i.e. even attempts are counted).
    """

    def __init__(
        self,
        game: Game,
        /,
        parent: Subscriber,
        key: str = None,
        limit: int = 1,
        *,
        only_successful: bool = False,
    ):
        if key is None:
            key = self.generate_key(parent)
        self._key = key
        self._limit = int(limit)
        self._only_successful = bool(only_successful)
        super().__init__(game, parent)
        CounterPerPhaseAux.get_or_create(game, key=self._key)

    @classmethod
    def generate_key(cls, parent: Subscriber) -> str:
        cn = cls.__qualname__
        return f"{cn}_{uuid4()}".replace("-", "")

    @property
    def key(self) -> str:
        return self._key

    @property
    def limit(self) -> int:
        return self._limit

    @property
    def only_successful(self) -> bool:
        return self._only_successful

    @property
    def counter(self) -> CounterPerPhaseAux:
        cppa = CounterPerPhaseAux.get_or_create(self.game, key=self._key)
        return cppa

    def hook_pre_action(self, action: Action) -> Optional[List[Action]]:
        if not self.only_successful:
            self.counter.increment()

    def hook_post_action(self, action: Action) -> Optional[List[Action]]:
        if self.only_successful:
            self.counter.increment()

    def check(self, action: Action) -> Optional[Constraint.Violation]:
        v = self.counter.value
        if v >= self.limit:
            return self.Violation(
                f"Reached limit of {self.limit} (found {v}) for key {self.key!r}"
            )


class LimitPerPhaseActorConstraint(LimitPerPhaseKeyConstraint, ATConstraint):
    """Allow only N uses per phase for this actor.

    Attributes
    ----------
    game : Game
    parent : ATBase
    limit : int
        The maximum number of actions per phase. Default is 1.
    only_successful : bool
        Whether to count only successful tries towards the limit.
        Default is False (i.e. even attempts are counted).
    """

    def __init__(
        self,
        game: Game,
        /,
        parent: ATBase,
        limit: int = 1,
        *,
        only_successful: bool = False,
    ):
        key = self.generate_key(parent)
        super().__init__(
            game, parent, key, limit=limit, only_successful=only_successful
        )

    @classmethod
    def generate_key(cls, parent: ATBase) -> str:
        cn = cls.__qualname__
        return f"{cn}_{parent.owner.name}"

    @property
    def owner(self) -> Actor:
        return self.parent.owner


class ConstraintNoSelfTarget(ATConstraint):
    """Any targets for the action, if they are Actors, must not be the owner."""

    def check(self, action: Action) -> Optional[Constraint.Violation]:
        ai = ActionInspector(action)
        p2a: Dict[str, Actor] = ai.values_of_type(Actor)
        bads = []
        for p, a in p2a.items():
            if a == self.owner:
                bads.append(f"{p!r}")
        if len(bads) > 0:
            return self.Violation("Self-targeting not allowed: " + ", ".join(bads))


class ConstraintNoSelfFactionTarget(ATConstraint):
    """Any targets for the action, if they are Actors, must be from other factions."""

    def check(self, action: Action) -> Optional[Constraint.Violation]:
        ai = ActionInspector(action)
        p2a: Dict[str, Actor] = ai.values_of_type(Actor)
        bads = []
        for p, a in p2a.items():
            if any(f in self.owner.factions for f in a.factions):
                bads.append(f"{p!r} ({a.name!r})")
        s = "" if len(self.owner.factions) == 1 else "s"
        if len(bads) > 0:
            return self.Violation(f"Targets in own faction{s}: " + ", ".join(bads))



import inspect
from typing import Callable, Type

from open_mafia_engine.util.repr import ReprMixin


class MafiaError(Exception, ReprMixin):
    """Base class for Mafia exceptions."""


class MafiaAmbiguousTypeName(MafiaError):
    """The type name conficts with an existing name."""

    def __init__(self, existing_type: Type[object], new_type: Type[object]) -> None:
        self.existing_type = existing_type
        self.new_type = new_type
        self.type_name = type_name = existing_type.__qualname__
        
        super().__init__(
            f"""Type {type_name!r} conficts with existing type.
            Existing type defined in: {inspect.getmodule(existing_type)}
            New type defined in: {inspect.getmodule(new_type)}            
            """
        )


class MafiaTypeNotFound(MafiaError):
    """The type was not found."""

    def __init__(self, type_name: str) -> None:
        self.type_name = type_name
        super().__init__(f"Couldn't find GameObject subtype {type_name!r}")


class MafiaConverterError(MafiaError, TypeError):
    """Could not convert object to the requested type."""

    def __init__(self, obj: str, type_: Type):
        self.obj = obj
        self.type_ = type_
        super().__init__(f"Couldn't convert {obj!r} to {type_!r}")


class MafiaBadHandler(MafiaError, TypeError):
    """Function can't be used as an event handler."""

    def __init__(self, func: Callable):
        self.func = func
        super().__init__(f"Function isn't a legal event handler: {func!r}")


class MafiaBadBuilder(MafiaError, TypeError):
    """Function can't be used as a game builder."""

    def __init__(self, func: Callable):
        self.func = func
        super().__init__(f"Function isn't a legal game builder: {func!r}")



from typing import List

from open_mafia_engine.core.all import (
    Actor,
    Event,
    Faction,
    Game,
)


class ENotify(Event):
    """Base event purely for notifications."""

    def __init__(self, game: Game, /, message: str):
        self.message = message
        super().__init__(game)


class EFActorNotify(ENotify):
    """Event that causes a message to be sent to an Actor."""

    def __init__(self, game: Game, /, message: str, actor: Actor):
        super().__init__(game, message=message)
        self.actor = actor


class EFactionNotify(ENotify):
    """Event that causes a message to be sent to a faction."""

    def __init__(self, game: Game, /, message: str, faction: Faction):
        super().__init__(game, message=message)
        self.faction = faction

    @property
    def actors(self) -> List[Actor]:
        return self.faction.actors



from __future__ import annotations

import inspect
import logging
from textwrap import indent
import warnings
from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    Union,
)

from makefun import with_signature

from open_mafia_engine.core.event_system import (
    Action,
    ActionInspector,
    Constraint,
    Event,
    Subscriber,
    handler,
)
from open_mafia_engine.core.game_object import GameObject, inject_converters
from open_mafia_engine.core.naming import ABILITY, TRIGGER, get_path

# from open_mafia_engine.core.outcome import Outcome, OutcomeAction

if TYPE_CHECKING:
    from open_mafia_engine.core.game import Game

logger = logging.getLogger(__name__)


class Faction(GameObject):
    """Faction, a.k.a. Alignment."""

    def __init__(self, game: Game, /, name: str):
        self.name = name
        self._actors: List[Actor] = []
        self._outcome_checkers: List[OutcomeChecker] = []
        super().__init__(game)

    @property
    def actors(self) -> List[Actor]:
        return list(self._actors)

    @property
    def actor_names(self) -> List[str]:
        return [x.name for x in self._actors]

    @property
    def outcome_checkers(self) -> List[OutcomeChecker]:
        return list(self._outcome_checkers)

    @inject_converters
    def add_actor(self, actor: Actor):
        if not isinstance(actor, Actor):
            raise TypeError(f"Expected Actor, got {actor!r}")
        if actor in self._actors:
            return
        self._actors.append(actor)
        actor._factions.append(self)

    @inject_converters
    def remove_actor(self, actor: Actor):
        if actor in self._actors:
            self._actors.remove(actor)
        if self in actor._factions:
            actor._factions.remove(self)

    def add_outcome_checker(self, oc: OutcomeChecker):
        if not isinstance(oc, OutcomeChecker):
            raise TypeError(f"Expected OutcomeChecker, got {oc!r}")
        if oc in self._outcome_checkers:
            return
        self._outcome_checkers.append(oc)
        oc._parent = self

    # def remove_outcome_checker(self, oc: OutcomeChecker):
    #     if oc in self._outcome_checkers:
    #         self._outcome_checkers.remove(self)
    #         # Wait, what do we do with `oc._parent`?!


class OutcomeChecker(Subscriber):
    """Checks for Faction Outcome. Base class."""

    def __init__(
        self, game: Game, parent: Faction, /, *, use_default_constraints: bool = True
    ):
        super().__init__(game, use_default_constraints=use_default_constraints)
        self._parent = parent
        parent.add_outcome_checker(self)

    @property
    def parent(self) -> Faction:
        return self._parent


class Actor(GameObject):
    """Actor object."""

    def __init__(self, game: Game, /, name: str, status: Dict[str, Any] = None):
        if status is None:
            status = {}
        self.name = name
        self._abilities: List[Ability] = []
        self._triggers: List[Trigger] = []
        self._factions: List[Faction] = []
        self._status: Status = Status(game, self, **status)
        super().__init__(game)

    @property
    def status(self) -> Status:
        return self._status

    @property
    def abilities(self) -> List[Ability]:
        return list(self._abilities)

    @property
    def ability_names(self) -> List[str]:
        return [a.name for a in self._abilities]

    @property
    def triggers(self) -> List[Trigger]:
        return list(self._triggers)

    @property
    def trigger_names(self) -> List[str]:
        return [t.name for t in self._triggers]

    @property
    def factions(self) -> List[Faction]:
        return list(self._factions)

    def add_ability(self, ability: Ability):
        """Adds this ability to self, possibly removing the old owner."""
        if not isinstance(ability, Ability):
            raise TypeError(f"Expected Ability, got {ability!r}")
        if ability in self._abilities:
            return
        self._abilities.append(ability)
        if ability._owner is not self:
            ability._owner._abilities.remove(ability)
            ability._owner = self

    def add_trigger(self, trigger: Trigger):
        """Adds this trigger to self, possibly removing the old owner."""
        if not isinstance(trigger, Trigger):
            raise TypeError(f"Expected Trigger, got {trigger!r}")
        if trigger in self._triggers:
            return
        self._triggers.append(trigger)
        if trigger._owner is not self:
            trigger._owner._triggers.remove(trigger)
            trigger._owner = self

    def add(self, obj: Union[Ability, Trigger, Faction]):
        """Adds """
        if isinstance(obj, Ability):
            self.add_ability(obj)
        elif isinstance(obj, Trigger):
            self.add_trigger(obj)
        elif isinstance(obj, Faction):
            warnings.warn("Better to do `Faction.add_actor(me)` directly.")
            obj.add_actor(self)
        else:
            raise TypeError(f"Expected Ability or Trigger, got {obj!r}")


class ATBase(Subscriber):
    """Base object for abilities and triggers.

    Attributes
    ----------
    game
    owner : Actor
    name : str
        The name of this object.
    desc : str
        Description. Default is "".
    """

    def __init__(
        self,
        game: Game,
        /,
        owner: Actor,
        name: str,
        desc: str = "",
        *,
        use_default_constraints: bool = True,
    ):
        if not isinstance(owner, Actor):
            raise TypeError(f"Expected Actor, got {owner!r}")
        if desc is None:
            desc = ""

        self._owner = owner
        self._name = str(name)
        self._desc = str(desc)
        super().__init__(game, use_default_constraints=use_default_constraints)
        owner.add(self)

    @property
    def owner(self) -> Actor:
        return self._owner

    @property
    def name(self) -> str:
        return self._name

    @property
    def desc(self) -> str:
        return self._desc

    def add_default_constraints(self):
        super().add_default_constraints()
        # Default constraints for abilities and triggers:
        ConstraintOwnerAlive(self.game, self)
        ConstraintActorTargetsAlive(self.game, self)

    @property
    def prefix_tags(self) -> List[str]:
        """These are used for descriptions."""
        res = []
        for con in self.constraints:
            res += con.prefix_tags
        return res

    @property
    def argument_names(self) -> List[str]:
        """Names of arguments used when calling. Default is empty."""
        return []

    def full_description(
        self, full_constraints: bool = False, highlight_method: Callable = None
    ) -> str:
        """Gets the full description of the Action/Trigger.

        If `full_constraints`, also adds the full descriptions of all constraints.
        By default, though, it skips these, adding only the prefix tags.
        """
        prestr = ", ".join(self.prefix_tags)
        if prestr == "":
            prestr = "Any"
        arg_names = self.argument_names
        abil_targs = ", ".join(arg_names)
        if abil_targs != "":
            abil_targs = f" ({abil_targs})"

        if highlight_method is None:
            name = self.name
        else:
            name = highlight_method(self.name)

        res = f"[{prestr}] {name}{abil_targs}: {self.desc}"
        if full_constraints:
            fcd = []
            # TODO: Long descriptions for constraints
            fcd = [indent(x, ".   ") for x in fcd]
            res = "\n".join([res] + fcd)
        return res


class ATConstraint(Constraint):
    """Constraint for Actions and Triggers. Will raise if used elsewhere."""

    def __init__(self, game, /, parent: ATBase):
        if not isinstance(parent, ATBase):
            raise TypeError(f"Constraint only available for ATBase, got {parent!r}")
        super().__init__(game, parent)

    @property
    def parent(self) -> ATBase:
        return self._parent

    @property
    def owner(self) -> Actor:
        """The owner of this constraint's parent."""
        return self.parent.owner


# NOTE: These constraints are default, and pretty fundamental, which is why they're here


class ConstraintOwnerAlive(ATConstraint):
    """The ability's (trigger's) owner must be alive."""

    def check(self, action: Action) -> Optional[Constraint.Violation]:
        if self.owner.status["dead"]:  # default is None, which is falsey
            return self.Violation(f"{self.owner.name!r} (owner) is dead.")


class ConstraintActorTargetsAlive(ATConstraint):
    """Any targets for the action, if they are Actors, must be alive."""

    def check(self, action: Action) -> Optional[Constraint.Violation]:
        ai = ActionInspector(action)
        p2a: Dict[str, Actor] = ai.values_of_type(Actor)
        bads = []
        for p, a in p2a.items():
            if a.status["dead"]:  # default is None, which is falsy
                bads.append(f"{p!r} ({a.name!r})")
        if len(bads) > 0:
            return self.Violation("Targets are dead: " + ", ".join(bads))


class Trigger(ATBase):
    """Basic Trigger object.

    Attributes
    ----------
    game : Game
    owner : Actor
    name : str
        The Trigger's name.
    desc : str
        Description. Default is "".
    """

    @property
    def path(self) -> str:
        return get_path(self.owner.name, TRIGGER, self.name)

    @property
    def prefix_tags(self) -> List[str]:
        return ["Auto"] + super().prefix_tags


class EActivate(Event):
    """Event of ability activation.

    That is, this event is triggered by a player trying to activate their Ability.
    """

    def __init__(self, game: Game, ability: Ability, /, *args, **kwargs):
        self._ability = ability
        self._args = args
        self._kwargs = kwargs
        super().__init__(game)

    @property
    def ability(self) -> Ability:
        return self._ability

    @property
    def args(self) -> Tuple:
        return self._args

    @property
    def kwargs(self) -> Dict[str, Any]:
        return dict(self._kwargs)


class Ability(ATBase):
    """Basic Ability object.

    Attributes
    ----------
    game
    owner : Actor
    name : str
        The Ability's name.
    desc : str
        Description. Default is "".
    """

    @property
    def argument_names(self) -> List[str]:
        """Names of arguments. User-facing."""
        sig = inspect.signature(self.activate)
        res = []
        for name, p in sig.parameters.items():
            if p.kind in [p.VAR_KEYWORD, p.VAR_POSITIONAL]:
                res.append("...")
            else:
                res.append(name)
        return res

    @property
    def path(self) -> str:
        return get_path(self.owner.name, ABILITY, self.name)

    @abstractmethod
    def activate(self, *args, **kwargs) -> Optional[List[Action]]:
        """Activate this ability with some arguments.

        Make the signature be the same as the Action's.
        If generated, it should match already.
        """

    @handler
    def handle_activate(self, event: EActivate) -> Optional[List[Action]]:
        """Handler to activate this ability."""
        if isinstance(event, EActivate) and event.ability is self:
            return self.activate(*event.args, **event.kwargs)
        return None

    @classmethod
    def generate(
        cls,
        action_or_func: Union[Type[Action], Callable],
        params: List[str] = None,
        name: str = None,
        doc: str = None,
        desc: str = None,
    ) -> Type[Ability]:
        """Create an Ability subtype from an Action or function.

        Parameters
        ----------
        action_or_func : Type[Action] or Callable
            If an Action subclass, uses it directly.
            If a callable (function), generates an action type and uses it.
        params : List[str]
            The names of parameters to leave as activation parameters.
            The rest will be taken as arguments for the Ability itself.
        """

        # Create an action type
        if isinstance(action_or_func, type) and issubclass(action_or_func, Action):
            TAction = action_or_func
        elif callable(action_or_func):
            gen_name: str = name if name else f"{action_or_func.__name__}_Action"
            TAction = Action.generate(action_or_func, name=gen_name, doc=doc)
        else:
            raise TypeError(f"Expected Action or function, got {action_or_func!r}")
        TAction: Type[Action]

        # Fix input arguments
        if params is None:
            params = []
        if name is None:
            name = f"{cls.__name__}_{TAction.__name__}"
            # TODO - add random bits to avoid conflict?
        if doc is None:
            doc = "(GENERATED ABILITY) " + (TAction.__doc__ or "")

        # Split the signature into __init__() and activate() args
        sig_action = inspect.signature(TAction.__init__)
        par_action = list(sig_action.parameters.values())

        par_activate = [
            inspect.Parameter("self", kind=inspect.Parameter.POSITIONAL_OR_KEYWORD)
        ] + [p for p in par_action if p.name in params]
        sig_activate = inspect.Signature(
            par_activate, return_annotation=Optional[List[Action]]
        )

        sig_init = "TODO"  # TODO: Actually create a proper init signature!

        # TODO: Set sig_init
        def __init__(self, game: Game, /, owner: Actor, name: str, desc: str = desc):
            super().__init__(game, owner, name, desc=desc)

        @with_signature(sig_activate)
        def activate(self, *args, **kwargs) -> Optional[List[Action]]:
            """Activate this ability with some arguments."""

            try:
                return [self.TAction(self.game, self, *args, **kwargs)]
            except Exception as e:
                logger.exception("Error executing action:")
                if False:  # Set for debugging errors :)
                    raise
                return None

        GeneratedAbility = type(
            name, (cls,), {"__init__": __init__, "activate": activate, "__doc__": doc}
        )
        GeneratedAbility.TAction = TAction

        return GeneratedAbility


class Status(GameObject, MutableMapping):
    """dict-like representation of an actor's status.

    Access of empty attribs gives None.
    Changing an attribute emits an EStatusChange event.

    Attributes
    ----------
    parent: Actor
    attribs : dict
        Raw keyword arguments for the status.
    """

    def __init__(self, game, /, parent: Actor, **attribs: Dict[str, Any]):
        super().__init__(game)
        self._parent = parent
        self._attribs = attribs

    @property
    def parent(self) -> Actor:
        return self._parent

    @property
    def attribs(self) -> Dict[str, Any]:
        return dict(self._attribs)

    def __getitem__(self, key) -> Any:
        return self._attribs.get(key, None)

    def __delitem__(self, key) -> None:
        old_val = self[key]
        if old_val is not None:
            del self._attribs[key]
        if old_val is None:
            return
        self.game.process_event(
            EStatusChange(self.game, self.parent, key, old_val, None)
        )

    def __setitem__(self, key, value) -> None:
        old_val = self[key]
        self._attribs[key] = value
        if old_val == value:
            return
        self.game.process_event(
            EStatusChange(self.game, self.parent, key, old_val, value)
        )

    def __len__(self) -> int:
        return len(self._attribs)

    def __iter__(self):
        return iter(self._attribs)

    def __repr__(self):
        cn = type(self).__qualname__
        parts = [repr(self.game), repr(self.parent)]
        for k, v in self._attribs.items():
            parts.append(f"{k}={v!r}")
        return f"{cn}(" + ", ".join(parts) + ")"


class EStatusChange(Event):
    """The Status has changed for some Actor."""

    def __init__(self, game, /, actor: Actor, key: str, old_val: Any, new_val: Any):
        super().__init__(game)
        self._actor = actor
        self._key = key
        self._old_val = old_val
        self._new_val = new_val

    @property
    def actor(self) -> Actor:
        return self._actor

    @property
    def status(self) -> Status:
        return self._actor.status

    @property
    def key(self) -> str:
        return self._key

    @property
    def old_val(self) -> Any:
        return self._old_val

    @property
    def new_val(self) -> Any:
        return self._new_val



import inspect
from reprlib import recursive_repr


class ReprMixin(object):
    """This mixin gives a reasonably smart automatic __repr__.

    It hasn't been extensively tested - might fail, so test yourself!"""

    @recursive_repr()
    def __repr__(self):
        sig = inspect.signature(self.__init__)
        # NOTE: str(sig) gives the raw signature, but without self
        used_keys = []
        s_args = []
        for k, param in sig.parameters.items():
            if param.kind == inspect.Parameter.POSITIONAL_ONLY:
                try:
                    val = getattr(self, k)
                    if val != param.default:
                        s_args.append(repr(val))
                except AttributeError:
                    s_args.append("<?>")
            elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                # *args
                try:
                    val = getattr(self, k)
                    for v in val:
                        s_args.append(repr(v))
                except AttributeError:
                    s_args.append("...")
            elif param.kind == inspect.Parameter.VAR_KEYWORD:
                # **kwargs
                dct = getattr(self, "__dict__", {})
                for k, v in dct.items():
                    if k not in used_keys:
                        try:
                            val = getattr(self, k)
                            s_args.append(f"{k}={repr(val)}")
                        except AttributeError:
                            s_args.append(f"{k}=<?>")
                        used_keys.append(k)
            elif param.kind in [
                inspect.Parameter.KEYWORD_ONLY,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
            ]:
                # keyword-able
                try:
                    val = getattr(self, k)
                    if val != param.default:  # could be 'empty'
                        s_args.append(f"{k}={repr(val)}")
                except AttributeError:
                    s_args.append(f"{k}=<?>")
                used_keys.append(k)
        res = f"{self.__class__.__qualname__}(" + ", ".join(s_args) + ")"
        return res



from __future__ import annotations

from pathlib import Path
from typing import Callable, IO, List, Optional, Union

import cloudpickle

from open_mafia_engine.core.auxiliary import AuxHelper, AuxObject
from open_mafia_engine.core.enums import ActionResolutionType
from open_mafia_engine.core.event_system import Action, ActionQueue, Event, EventEngine
from open_mafia_engine.core.game_object import GameObject
from open_mafia_engine.core.phase_cycle import (
    AbstractPhaseSystem,
    ETryPhaseChange,
    Phase,
    PhaseChangeAction,
    SimplePhaseCycle,
)
from open_mafia_engine.core.state import Actor, Faction


class Game(object):
    """Defines the state of an entire game, including the execution context.

    Parameters
    ----------
    gen_phases : Callable[[Game], AbstractPhaseSystem]
        A function that creates a phase system, given a Game.
        Use `AbstractPhaseSystem.gen(*args, **kwargs)` to create this function.
        The default is `SimplePhaseCycle.gen()`, which creates a day-night cycle.

    Attributes
    ----------
    event_engine : EventEngine
        The event engine, which handles all subscription and broadcasting.
    action_queue : ActionQueue
        The base action queue, which holds all core-level actions & all history.
    phase_system : AbstractPhaseSystem
        System that defines phases and their transitions.
    actors : List[Actor]
        All the actors (players) of the game.
    factions : List[Faction]
        All the factions (teams) of the game.
    aux : AuxHelper
        Helper object for auxiliary game objects.
    """

    def __init__(
        self,
        gen_phases: Callable[[Game], AbstractPhaseSystem] = SimplePhaseCycle.gen(),
    ):
        self._event_engine = EventEngine(self)
        self._action_queue = ActionQueue(self)
        self._actors: List[Actor] = []
        self._factions: List[Faction] = []
        self._phase_system: AbstractPhaseSystem = gen_phases(self)
        self._aux = AuxHelper(self)

    def __repr__(self):
        cn = type(self).__qualname__
        return f"{cn}()"

    @property
    def event_engine(self) -> EventEngine:
        return self._event_engine

    @property
    def action_queue(self) -> ActionQueue:
        return self._action_queue

    @property
    def actors(self) -> List[Actor]:
        return list(self._actors)

    @property
    def actor_names(self) -> List[str]:
        return [x.name for x in self._actors]

    @property
    def factions(self) -> List[Faction]:
        return list(self._factions)

    @property
    def faction_names(self) -> List[str]:
        return [x.name for x in self._factions]

    @property
    def phase_system(self) -> AbstractPhaseSystem:
        return self._phase_system

    @property
    def current_phase(self) -> Phase:
        return self.phase_system.current_phase

    @property
    def aux(self) -> AuxHelper:
        return self._aux

    def add(self, obj: GameObject):
        """Adds the object to this game.

        This is automatically called during `obj.__init__()`
        """
        if isinstance(obj, Actor):
            if obj not in self._actors:
                self._actors.append(obj)
        elif isinstance(obj, Faction):
            if obj not in self._factions:
                self._factions.append(obj)
        elif isinstance(obj, AuxObject):
            self._aux.add(obj)
        # NOTE: We ignore all other objects, but don't throw.

    # TODO: remove()?

    def process_event(self, event: Event, *, process_now: bool = False):
        """Processes the action."""
        responses: List[Action] = self.event_engine.broadcast(event)
        for resp in responses:
            self.action_queue.enqueue(resp)

        process_now = (
            process_now
            or self.current_phase.action_resolution == ActionResolutionType.instant
            or isinstance(event, ETryPhaseChange)
        )
        if process_now:
            self.action_queue.process_all()

    def change_phase(self, new_phase: Optional[Phase] = None):
        """Changes the phase to the given one (or bumps it). This causes events."""
        # NOTE: You can pass strings here, because ETryPhaseChange.__init__ converts
        self.process_event(ETryPhaseChange(self, new_phase=new_phase))
        # `process_now=True` is implicit :)

    def save(self, file: Union[Path, str, IO[bytes]]):
        """Cloudpickle-based saving"""
        if isinstance(file, Path):
            with file.open(mode="wb") as f:
                self.save(f)
        elif isinstance(file, str):
            return self.save(Path(file))
        else:
            cloudpickle.dump(self, file)

    @classmethod
    def load(cls, file: Union[Path, str, IO[bytes]]) -> Game:
        """Cloudpickle-based loading"""

        if isinstance(file, Path):
            with file.open(mode="rb") as f:
                return cls.load(f)
        elif isinstance(file, str):
            return cls.load(Path(file))
        else:
            res = cloudpickle.load(file)
            if not isinstance(res, cls):
                raise TypeError(f"Wrong object was pickled: {file!r}")
            return res


from enum import Enum


class ActionResolutionType(str, Enum):
    """How actions are resolved."""

    instant = "instant"
    end_of_phase = "end_of_phase"

    def __repr__(self):
        cn = type(self).__qualname__
        return f"{cn}({self.value!r})"


class Outcome(str, Enum):
    """Outcome (victory or defeat)."""

    victory = "victory"
    defeat = "defeat"

    def __repr__(self):
        cn = type(self).__qualname__
        return f"{cn}({self.value!r})"



"""Example command-line application built using prompt_toolkit and open_mafia_engine."""

try:
    import prompt_toolkit
except ImportError:
    raise ImportError(
        "Please install prompt_toolkit, either directly or via extras:\n"
        "pip install open_mafia_engine[examples]",
        name="prompt_toolkit",
    )

import shlex
import traceback
from textwrap import indent

from prompt_toolkit.application import Application
from prompt_toolkit.application.current import get_app
from prompt_toolkit.completion import FuzzyWordCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.layout.containers import HSplit, VSplit, Window
from prompt_toolkit.layout.layout import Layout
from prompt_toolkit.styles import Style
from prompt_toolkit.widgets import SearchToolbar, TextArea

import open_mafia_engine.api as mafia

if mafia.PYGMENTS_AVAILABLE:
    from prompt_toolkit.lexers import PygmentsLexer

    lexer = PygmentsLexer(mafia.MafiaCliLexer)
else:
    lexer = None

help_str = """
Non-mafia commands:
  exit - exit the application (also works with Ctrl-C twice, or Ctrl-Q twice)
  help - show this help

Mafia command structure:
  USER COMMAND [ARG1 ...]

Basic commands:
  USER join               Joins the game.
  USER out                Leaves the game.
  ADMIN create-game test  Creates the "test game".
  ADMIN phase             Change the phase.
  USER vote TARGET        Votes for the TARGET.
  USER do action ARG      Does an action, with one or more arguments.
"""

# Create the Mafia command runner, which controls the lobby and game
runner = mafia.CommandRunner[str](
    parser=mafia.ShellCommandParser(),
    lobby=mafia.AutoAddStrLobby(admin_names=["admin"]),
)

# Add key bindings.
kb = KeyBindings()


@kb.add("c-c", "c-c")
@kb.add("c-q", "c-q")
async def _(event):
    "Pressing Ctrl-Q or Ctrl-C twicw will exit the user interface."

    get_app().exit()


history_field = TextArea(
    style="class:history-field",
    text="",
    lexer=lexer,
)
status_field = TextArea(
    style="class:status-field",
    text="Current status will appear here.",
)


def set_status_text(msg: str):
    """Sets the status field text."""
    new_text = "Current Status: " + msg
    status_field.buffer.document = Document(
        text=new_text, cursor_position=len(new_text)
    )


SEP = "-----------------"


def update_status_text():
    admstr = ", ".join(runner.lobby.admin_names)
    if runner.in_game:
        game: mafia.Game = runner.game

        # Game status
        txt = f"[In Game]\n\nAdmins: {admstr}\nPhase: {game.current_phase.name}\n"

        # Actor status
        txt += "\nActor Status:"
        for act in game.actors:
            txt += (
                f"\n\n{SEP}\n"
                + f"{act.name} - {', '.join(f.name for f in act.factions)}"
                + "\n"
                + indent(
                    "\n".join([abil.full_description() for abil in act.abilities]),
                    "  ",
                )
            )
            if len(act.status) > 0:
                txt += "\nStatus:\n" + "\n".join(
                    [f"  {k}: {v}" for k, v in act.status.items()]
                )
        txt += f"\n\n{SEP}\n"

        # Display vote tally status
        tallies = game.aux.filter_by_type(mafia.LynchTally)
        if len(tallies) == 1:
            tally: mafia.LynchTally = tallies[0]

            vr = tally.results
            if len(vr.vote_counts) > 0:
                vres = ["Vote Count:"]
                # TODO: Make sure this is proper who-votes-for-whom behavior.
                for go, cnt, voters in vr.vote_map:
                    if cnt <= 0:
                        continue
                    if isinstance(go, mafia.Actor):
                        name = go.name
                    elif isinstance(go, mafia.VoteAgainstAll):
                        name = "No Lynch"
                    else:
                        name = str(go)
                    voter_str = ", ".join([v.name for v in voters])
                    vres.append(f"  {name} ({cnt}) - {voter_str}")

                try:
                    # This is less tested, but hopefully won't fail
                    vres.append("")
                    vres.append("Vote Leaders:")
                    vls = []
                    for vl in vr.vote_leaders:
                        if isinstance(vl, mafia.Actor):
                            vls.append(vl.name)
                        elif isinstance(vl, mafia.VoteAgainstAll):
                            vls.append("No Lynch")
                        else:
                            vls.append(str(vl))
                    vres.append("  " + ", ".join(vls))
                except Exception:
                    pass
                txt += "\n" + "\n".join(vres)
    else:
        txt = f"[In Lobby]\n\nAdmins: {admstr}\nPlayers:\n"
        if len(runner.lobby.players) == 0:
            txt += "  <no players>"
        else:
            txt += "\n".join(f"  {x}" for x in runner.lobby.player_names)
    set_status_text(txt)


def get_words():
    """Gets possible words for completion.

    NOTE: This runs and is called, but is ignored for some reason.
    """
    res = (
        ["exit", "help"]
        + runner.currently_available_commands(source="admin")
        + runner.lobby.admin_names
        + runner.lobby.player_names
    )
    if runner.in_game:
        res += runner.game.faction_names
        res += [x.name for x in runner.game.phase_system.possible_phases]
    return res


search_field = SearchToolbar()
input_field = TextArea(
    height=1,
    prompt=">>> ",
    style="class:input-field",
    multiline=False,
    wrap_lines=False,
    search_field=search_field,
    lexer=lexer,
    completer=FuzzyWordCompleter(get_words),
)


def submit_command(buff):
    """This runs when we press 'enter' on the input field."""
    text = input_field.text

    # Add text to output buffer.
    new_history = history_field.text + "\n> " + text + "\n"

    # Handle with the runner
    _parts = shlex.split(text, posix=True)
    if len(_parts) == 0:
        return
    src = _parts[0]
    other = ", ".join([f'"{x}"' for x in _parts[1:]])

    if src == "exit":
        get_app().exit()
    elif src == "help":
        new_history += help_str + "\n"
    else:
        try:
            runner.parse_and_run(src, other)
            # TODO: How do we return stuff?
            update_status_text()
        except Exception as e:
            err_str = traceback.format_exc()
            new_history += err_str + "\n"

    # Print text to history buffer
    history_field.buffer.document = Document(
        text=new_history, cursor_position=len(new_history)
    )


input_field.accept_handler = submit_command

container = HSplit(
    [
        VSplit(
            [status_field, Window(width=1, char="|", style="class:line"), history_field]
        ),
        Window(height=1, char="-", style="class:line"),
        input_field,
        search_field,
    ]
)


# Define application.
style = Style(
    [
        ("input-field", "bg:#000000 #ffffff"),
        ("history-field", "bg:#000000 #ffffff"),
        ("status-field", "bg:#000000 #ffffff"),
        ("line", "#004400"),
    ]
)
application = Application(
    layout=Layout(container, focused_element=input_field),
    key_bindings=kb,
    style=style,
    # mouse_support=False,
    full_screen=True,
)


def main():
    # TODO: Add Typer arguments to load external modules.
    update_status_text()
    application.run()


if __name__ == "__main__":
    main()



from open_mafia_engine.core.all import PhaseChangeAction, Ability


PhaseChangeAbility = Ability.generate(
    PhaseChangeAction,
    params=["new_phase"],
    name="PhaseChangeAbility",
    doc="Ability to change the phase. Typically admin-only.",
    desc="Change the phase (or just bump it).",
)



"""API package.

Use as the following:

    import open_mafia_engine.api as mafia
    mafia.Game
"""

from open_mafia_engine.core.all import *
from open_mafia_engine.built_in.all import *
from open_mafia_engine.builders.all import *
from open_mafia_engine.commands.all import *



from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Mapping

from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from .raw import TUser

__all__ = ["AbstractLobby", "SimpleDictLobby", "AutoAddStrLobby"]


class AbstractLobby(Mapping, Generic[TUser]):
    """A lobby, with mappings to user-defined objects.

    Attributes
    ----------
    admins : dict
        Mapping of admin names to user objects. Override this.
    players : dict
        Mapping of player names to user objects. Override this.
    """

    @property
    @abstractmethod
    def players(self) -> Dict[str, TUser]:
        """Return mapping of str to players."""

    @property
    @abstractmethod
    def admins(self) -> Dict[str, TUser]:
        """Return mapping of str to admins."""

    @abstractmethod
    def add_player(self, name: str, user: TUser):
        """Add the player to the game."""

    @abstractmethod
    def remove_player(self, name: str, user: TUser):
        """Remove the player from the game."""

    @abstractmethod
    def add_admin(self, name: str, user: TUser):
        """Add the admin to the game."""

    @abstractmethod
    def remove_admin(self, name: str, user: TUser):
        """Remove the admin from the game."""

    @property
    def player_names(self) -> List[str]:
        return list(self.players.keys())

    @property
    def admin_names(self) -> List[str]:
        return list(self.admins.keys())

    @property
    def all_names(self) -> List[str]:
        return list(set(self.player_names).union(self.admin_names))

    def __getitem__(self, k: str) -> TUser:
        res = self.admins.get(k, None)
        if res is None:
            res = self.players.get(k, None)
        if res is None:
            raise KeyError(k)
        return res

    def __iter__(self):
        keys = set(self.all_names)
        return iter(keys)

    def __len__(self):
        keys = set(self.all_names)
        return len(keys)


class SimpleDictLobby(AbstractLobby[TUser]):
    """Simple lobby implementation."""

    def __init__(
        self,
        admins: Dict[str, TUser] = None,
        players: Dict[str, TUser] = None,
    ) -> None:
        if admins is None:
            admins = {}
        if players is None:
            players = {}
        self._admins = dict(admins)
        self._players = dict(players)

    @property
    def players(self) -> Dict[str, TUser]:
        return dict(self._players)

    @property
    def admins(self) -> Dict[str, TUser]:
        return dict(self._admins)

    def add_admin(self, name: str, user: TUser):
        self._admins[name] = user

    def remove_admin(self, name: str, user: TUser):
        if name in self._admins:
            del self._admins[name]

    def add_player(self, name: str, user: TUser):
        self._players[name] = user

    def remove_player(self, name: str, user: TUser):
        if name in self._players:
            del self._players[name]


class AutoAddStrLobby(AbstractLobby[str]):
    """String-based lobby that adds any users automatically."""

    def __init__(self, admin_names: List[str] = [], player_names: List[str] = []):
        self._admin_names = set(admin_names)
        self._player_names = set(player_names)

    @property
    def players(self) -> Dict[str, str]:
        return {x: x for x in self._player_names}

    @property
    def admins(self) -> Dict[str, str]:
        return {x: x for x in self._admin_names}

    @property
    def player_names(self) -> List[str]:
        return sorted(self.players.keys())

    @property
    def admin_names(self) -> List[str]:
        return sorted(self.admins.keys())

    def add_admin(self, name: str, user: str):
        assert name == user
        self._admin_names.add(name)

    def remove_admin(self, name: str, user: str):
        self._admin_names.discard(name)

    def add_player(self, name: str, user: str):
        assert name == user
        self._player_names.add(name)

    def remove_player(self, name: str, user: str):
        self._player_names.discard(name)

    def __getitem__(self, k: str) -> str:
        return k



from open_mafia_engine.core.all import (
    Action,
    Actor,
    CancelAction,
    ConditionalCancelAction,
    EPreAction,
    Trigger,
    handler,
)

from .kills import KillAction, LynchAction


class UnkillableTrigger(Trigger):
    """Cancels kill actions against the owner."""

    @handler
    def cancel_self_kills(self, event: EPreAction):
        if isinstance(event, EPreAction) and isinstance(event.action, KillAction):
            # NOTE: This does almost the same, but we need to check ourselves.
            # if event.action.target == self.owner:
            #     return [CancelAction(self.game, self, target=event.action)]

            def cond(action: KillAction) -> bool:
                """Cancel only if the final target is self."""
                return action.target == self.owner

            return [
                ConditionalCancelAction(
                    self.game, self, target=event.action, condition=cond
                )
            ]


class UnlynchableTrigger(Trigger):
    """Cancels lynch actions against the owner."""

    @handler
    def cancel_self_kills(self, event: EPreAction):
        if isinstance(event, EPreAction) and isinstance(event.action, LynchAction):
            # NOTE: This does almost the same, but we need to check ourselves.
            # if event.action.target == self.owner:
            #     return [CancelAction(self.game, self, target=event.action)]

            def cond(action: LynchAction) -> bool:
                """Cancel only if the final target is self."""
                return action.target == self.owner

            return [
                ConditionalCancelAction(
                    self.game, self, target=event.action, condition=cond
                )
            ]



from typing import List, Optional

from open_mafia_engine.core.all import (
    Ability,
    Action,
    Actor,
    ATBase,
    ConditionalCancelAction,
    EPreAction,
    Game,
    GameObject,
    handler,
)

from .auxiliary import TempPhaseAux
from .kills import KillAction


class ProtectorFromKillAux(TempPhaseAux):
    """Aux object that protects the target from kills."""

    def __init__(
        self,
        game: Game,
        /,
        target: Actor,
        key: str = None,
        *,
        use_default_constraints: bool = True,
    ):
        self._target = target
        super().__init__(
            game, key=None, use_default_constraints=use_default_constraints
        )

    @property
    def target(self) -> Actor:
        return self._target

    @handler
    def handle_to_save(
        self, event: EPreAction
    ) -> Optional[List[ConditionalCancelAction]]:
        """Cancels the action if it came from the target."""

        if isinstance(event, EPreAction) and isinstance(event.action, KillAction):
            # NOTE: This will protect only our target from kills, even if this
            # kill is redirected somewhere.

            def cond(action: KillAction) -> bool:
                """Cancel only if the final target is self."""
                return action.target == self.target

            return [
                ConditionalCancelAction(
                    self.game, self, target=event.action, condition=cond
                )
            ]


class ProtectFromKillAction(Action):
    """Action that protects the tarte from kills until the end of the phase."""

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        target: Actor,
        *,
        priority: float = 80,
        canceled: bool = False,
    ):
        self._target = target
        super().__init__(game, source, priority=priority, canceled=canceled)

    @property
    def target(self) -> Actor:
        return self._target

    def doit(self):
        ProtectorFromKillAux(self.game, target=self.target)


ProtectFromKillAbility = Ability.generate(
    ProtectFromKillAction,
    params=["target"],
    name="ProtectFromKillAbility",
    doc="Ability to protect from kills",
    desc="Protects the target from kills.",
)



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Callable, Iterable, List, Optional, Tuple, Union

from open_mafia_engine.core.enums import ActionResolutionType
from open_mafia_engine.core.event_system import (
    Action,
    EPostAction,
    EPreAction,
    Event,
    Subscriber,
    handler,
)
from open_mafia_engine.core.game_object import GameObject, converter

if TYPE_CHECKING:
    from open_mafia_engine.core.game import Game


class Phase(GameObject):
    """Represents a monolithic "phase" of action.

    Attributes
    ----------
    name : str
        The current phase name.
    action_resolution : ActionResolutionType
        One of {"instant", "end_of_phase"}
    """

    def __init__(
        self,
        game,
        /,
        name: str,
        action_resolution: str = "instant",
    ):
        super().__init__(game)

        self._name = name
        self._action_resolution = ActionResolutionType(action_resolution)

    @property
    def name(self) -> str:
        return self._name

    @property
    def action_resolution(self) -> ActionResolutionType:
        return self._action_resolution

    @action_resolution.setter
    def action_resolution(self, v: str):
        self._action_resolution = ActionResolutionType(v)

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Phase):
            return NotImplemented
        return (o.name == self.name) and (o.action_resolution == self.action_resolution)


class ETryPhaseChange(Event):
    """Try to change the phase."""

    def __init__(self, game, /, new_phase: Optional[Phase] = None):
        if not (new_phase is None or isinstance(new_phase, Phase)):
            new_phase = converter.convert(game, Phase, new_phase)
        self._new_phase = new_phase
        super().__init__(game)

    @property
    def new_phase(self) -> Optional[Phase]:
        return self._new_phase


class EPrePhaseChange(EPreAction):
    """Phase is about to change."""

    @property
    def action(self) -> PhaseChangeAction:
        return self._action

    @property
    def new_phase(self) -> Optional[Phase]:
        return self.action.new_phase

    @property
    def old_phase(self) -> Phase:
        return self.action.old_phase


class EPostPhaseChange(EPostAction):
    """Phase has changed."""

    @property
    def action(self) -> PhaseChangeAction:
        return self._action

    @property
    def new_phase(self) -> Phase:
        np = self.action.new_phase
        if np is None:
            np = self.game.current_phase
        return np

    @property
    def old_phase(self) -> Phase:
        return self.action.old_phase


class PhaseChangeAction(Action):
    """Action to change the phase.

    Parameters
    ----------
    new_phase : None or Phase
        The resulting phase. By default, `None` uses the next phase.
    old_phase : Phase
        The phase that this action was created in.
    """

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        new_phase: Optional[Phase] = None,
        *,
        priority: float = 0.0,
        canceled: bool = False,
    ):
        super().__init__(game, source, priority=priority, canceled=canceled)
        self.new_phase = new_phase
        self._old_phase = self.game.current_phase

    @property
    def old_phase(self) -> Phase:
        return self._old_phase

    def doit(self):
        if self.new_phase is None:
            self.game.phase_system.bump_phase()
        else:
            self.game.phase_system.current_phase = self.new_phase

    Pre = EPrePhaseChange
    Post = EPostPhaseChange


class AbstractPhaseSystem(Subscriber):
    """Interface for a phase system.

    It's possible to see all phases by using `game.phase_system.possible_phases`
    """

    def __init__(self, game: Game, /, *, use_default_constraints: bool = True):
        super().__init__(game, use_default_constraints=use_default_constraints)
        self._startup = Phase(game, name="startup", action_resolution="instant")
        self._shutdown = Phase(game, name="shutdown", action_resolution="instant")

    @property
    def startup(self) -> Phase:
        return self._startup

    @property
    def shutdown(self) -> Phase:
        return self._shutdown

    @property
    @abstractmethod
    def possible_phases(self) -> Iterable[Phase]:
        """Returns all possible phases (as a new iterable).

        If it is infinite, override __getitem__ as well!
        """

    def __getitem__(self, key: str) -> Phase:
        """Returns the phase with the given name.

        By default, iterates over all possible phases.
        """
        if not isinstance(key, str):
            raise TypeError(f"Expected key as str, got {key!r}")
        for p in self.possible_phases:
            if key == p.name:
                return p
        raise KeyError(key)

    @property
    @abstractmethod
    def current_phase(self) -> Phase:
        """Returns the current phase."""

    @current_phase.setter
    @abstractmethod
    def current_phase(self, v: Phase):
        """Sets the current phase."""

    @abstractmethod
    def bump_phase(self) -> Phase:
        """Updates the phase to use the next one, then returns the current one."""

    @classmethod
    def gen(cls, *args, **kwargs) -> Callable[[Game], AbstractPhaseSystem]:
        """Create a callable that generates a phase cycle."""

        def func_gen(game: Game) -> AbstractPhaseSystem:
            return cls(game, *args, **kwargs)

        return func_gen

    @handler
    def system_phase_change(
        self, event: ETryPhaseChange
    ) -> Optional[List[PhaseChangeAction]]:
        """Some external system asked for a phase change."""
        if not isinstance(event, ETryPhaseChange):
            return
        return [PhaseChangeAction(self.game, self, event.new_phase)]


class SimplePhaseCycle(AbstractPhaseSystem):
    """Simple phase cycle definition.

    Parameters
    ----------
    game : Game
    cycle : None or List[Tuple[str, ActionResolutionType]]
        The cycle definition. Submit pairs of (name, resolution_type).
        By default, uses `[("day", "instant"), ("night", "end_of_phase")]`
    """

    _STARTUP = -1
    _SHUTDOWN = -2

    def __init__(
        self,
        game: Game,
        /,
        cycle: List[Tuple[str, ActionResolutionType]] = None,
        current_phase: Optional[str] = None,
    ):
        super().__init__(game)
        if cycle is None:
            cycle = [("day", "instant"), ("night", "end_of_phase")]
        cphases = []
        names = set()
        for name, ar in cycle:
            if name in names:
                raise ValueError(f"Duplicate name {name!r} in {names}")
            ar = ActionResolutionType(ar)
            names.add(name)
            cphases.append(Phase(game, name=name, action_resolution=ar))
        self._cycle = cphases
        self._i = self._STARTUP
        if current_phase is not None:
            self.current_phase = current_phase

    @property
    def cycle(self) -> List[Phase]:
        return list(self._cycle)

    @property
    def possible_phases(self) -> List[Phase]:
        return [self.startup, *self.cycle, self.shutdown]

    @property
    def current_phase(self) -> Phase:
        """Returns the current phase."""
        if self._i == self._STARTUP:
            return self.startup
        elif self._i == self._SHUTDOWN:
            return self.shutdown
        i = self._i % len(self._cycle)
        return self._cycle[i]

    @current_phase.setter
    def current_phase(self, v: Union[str, Phase]):
        if isinstance(v, str):
            new_phase = self[v]
        else:
            new_phase = v
        if new_phase == self.startup:
            # Maybe disallow going back to startup?
            self._i = self._STARTUP
        elif new_phase == self.shutdown:
            self._i = self._SHUTDOWN
        elif new_phase in self.possible_phases:
            # Just move through all phases implicitly - we won't trigger anything
            while self.current_phase != new_phase:
                self._i += 1
        else:
            raise ValueError(f"No such phase found: {v!r}")

    def bump_phase(self) -> Phase:
        """Updates the phase to use the next one, then returns the current one.

        Trying to bump on `shutdown` phase will be ignored.
        """
        if self._i == self._STARTUP:
            self._i = 0
            return self.current_phase
        elif self._i == self._SHUTDOWN:
            return self.current_phase
            # raise ValueError(f"Cannot bump shutdown phase: {self.shutdown}")
        self._i += 1
        return self.current_phase

    @classmethod
    def gen(
        cls,
        cycle: List[Tuple[str, ActionResolutionType]] = None,
        current_phase: Optional[str] = None,
    ) -> Callable[[Game], SimplePhaseCycle]:
        """Generator for a simple phase cycle."""
        return super().gen(cycle=cycle, current_phase=current_phase)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['open_mafia_engine',
 'open_mafia_engine.api',
 'open_mafia_engine.builders',
 'open_mafia_engine.built_in',
 'open_mafia_engine.commands',
 'open_mafia_engine.core',
 'open_mafia_engine.examples',
 'open_mafia_engine.test',
 'open_mafia_engine.util']

package_data = \
{'': ['*']}

install_requires = \
['cloudpickle>=1.6.0,<2.0.0',
 'fuzzywuzzy>=0.18.0,<0.19.0',
 'makefun>=1.11.3,<2.0.0',
 'sortedcontainers>=2.3.0,<3.0.0',
 'typer>=0.3.2,<0.4.0']

extras_require = \
{'dev': ['pytest>=6.2.4,<7.0.0', 'flake8>=3.9.2,<4.0.0', 'mypy>=0.800,<0.801'],
 'docs': ['pygments>=2.9.0,<2.10.0',
          'mkdocs>=1.1.2,<2.0.0',
          'mkdocs-material>=7.1.8,<8.0.0',
          'mkdocstrings>=0.15.2,<0.16.0',
          'pymdown-extensions>=8.2,<9.0'],
 'examples': ['prompt-toolkit>=3.0.19,<4.0.0', 'pygments>=2.9.0,<2.10.0'],
 'recommended': ['prompt-toolkit>=3.0.19,<4.0.0',
                 'pygments>=2.9.0,<2.10.0',
                 'pytest>=6.2.4,<7.0.0'],
 'test': ['pytest>=6.2.4,<7.0.0']}

entry_points = \
{'console_scripts': ['mafia-cli = open_mafia_engine.examples.cli:main']}

setup_kwargs = {
    'name': 'open-mafia-engine',
    'version': '0.5.0',
    'description': 'Open Mafia Engine - a framework for mafia/werewolf games.',
    'long_description': '# Open Mafia Engine\n\n[![PyPI version](https://badge.fury.io/py/open-mafia-engine.svg)](https://badge.fury.io/py/open-mafia-engine)\n[![Stable Release](https://img.shields.io/github/release/open-mafia/open_mafia_engine.svg)](https://github.com/open-mafia/open_mafia_engine/releases)\n[![Docs](https://readthedocs.org/projects/open-mafia-engine/badge/?version=latest)](https://open-mafia-engine.readthedocs.io/en/latest/)\n\n[![Test Python package "open_mafia_engine"](https://github.com/open-mafia/open_mafia_engine/actions/workflows/python-testing.yml/badge.svg)](https://github.com/open-mafia/open_mafia_engine/actions/workflows/python-testing.yml)\n\nThe Open Mafia Engine is a flexible, open-source game engine for Mafia-like games.\n\nSee the [Wikipedia article](https://en.wikipedia.org/wiki/Mafia_%28party_game%29)\nfor information on what Mafia/Werewolf is.\n\nThe Open Mafia Engine actually allows a much broader type of game to be played,\nwith custom roles and very intricate "setups". Most of these custom roles come\nfrom either "real-life" games or online forums such as\n[MafiaScum](https://wiki.mafiascum.net/index.php?title=Main_Page) and\n[Bay12Games Mafia Subforum](http://www.bay12forums.com/smf/index.php?board=20.0).\n\n## Features\n\n- Event-based architecture, which allows for very complex interactions.\n- Many built-in abilities, victory conditions, etc.\n  (Not "many" yet, but we\'re working on it!)\n- Games are defined declaratively or using an parametrized GameBuilder.\n- Open source & extensible.\n\n## Installing\n\nInstall the latest stable version, with recommended dependencies, via pip:\n\n`pip install open_mafia_engine[recommended]`\n\nSee the [dev docs](docs/development/installing_dev.md) to install for local\ndevelopment (using Poetry).\n\n## Getting started\n\nThis example starts a 5-player "test" mafia game (1 mafioso vs 4 townies):\n\n```python\nimport open_mafia_engine.api as mafia\n\nbuilder = mafia.GameBuilder.load("test")\nplayers = [\'Alice\', \'Bob\', \'Charlie\', \'Dave\', \'Eddie\']\ngame = builder.build(players)\n```\n\n## Example Application\n\nThis repository includes an example text-based Mafia app, runnable like:\n\n```bash\nmafia-cli\n# or\npython -m open_mafia_engine.example.cli\n```\n\nThis is what the UI looks like:\n\n![Example CLI Application](docs/examples/ExampleMafiaCLI.gif)\n',
    'author': 'Open Mafia Team',
    'author_email': 'openmafiateam@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://open-mafia-engine.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Mapping, Type, Union, get_args, get_origin
from uuid import uuid4

from open_mafia_engine.core.event_system import (
    Action,
    ActionQueue,
    Event,
    EventEngine,
    Subscriber,
)
from open_mafia_engine.core.game_object import GameObject

if TYPE_CHECKING:
    from open_mafia_engine.core.game import Game

NoneType = type(None)


class AuxObject(Subscriber):
    """Base class for auxiliary objects.

    Aux objects have a key they are retrievable by.

    Attributes
    ----------
    game : Game
    key : str
        The key to retrieve the object by. If None (default), generates one.
    """

    def __init__(
        self, game: Game, /, key: str = None, *, use_default_constraints: bool = True
    ):
        if key is None:
            key = self.generate_key()
        self._key = str(key)
        # NOTE: super().__init__ auto-adds self to the AuxHelper
        super().__init__(game, use_default_constraints=use_default_constraints)

    @classmethod
    def generate_key(cls) -> str:
        """Generates a key for this class (used if None is passed in __init__)."""
        return cls.__qualname__ + "_" + str(uuid4()).replace("-", "")

    @property
    def key(self) -> str:
        return self._key

    @classmethod
    def get_or_create(
        cls,
        game: Game,
        /,
        key: str = None,
        *,
        use_default_constraints: bool = True,
        **kwargs,
    ) -> AuxObject:
        """Finds the AuxObject by key or, if there is no object, creates it."""
        if key is None:
            key = cls.generate_key()
        res = game.aux.get(key)
        if res is None:
            res = cls(
                game, key=key, use_default_constraints=use_default_constraints, **kwargs
            )
        if not isinstance(res, cls):
            raise TypeError(f"Wrong type for AuxObject: expected {cls!r}, got {res!r}")
        return res


class AuxHelper(GameObject, Mapping):
    """Auxiliary object helper."""

    def __init__(self, game, /, max_objects: int = 1000):
        self._key_map: Dict[str, AuxObject] = {}
        # self._children: List[AuxObject] = []
        self._max_objects = int(max_objects)
        super().__init__(game)

    @property
    def max_objects(self) -> int:
        return self._max_objects

    @property
    def children(self) -> List[AuxObject]:
        return list(self._key_map.values())

    def __getitem__(self, key: str):
        return self._key_map[key]

    def __iter__(self):
        return iter(self._key_map)

    def __len__(self):
        return len(self._key_map)

    def add(self, obj: AuxObject):
        """Adds the aux object to self."""
        if not isinstance(obj, AuxObject):
            raise TypeError(f"Expected AuxObject, got {obj!r}")
        if self._key_map.get(obj.key) == obj:
            return
        elif obj.key in self._key_map.keys():
            # FIXME: different object with the same key!
            return
        elif obj in self._key_map.values():
            # FIXME: same object with different key? Key should be able to change.
            return
        # OK, let's add it
        if len(self) > self.max_objects:
            raise ValueError(f"Reached {self.max_objects} (max) aux objects!")
        self._key_map[obj.key] = obj

    def remove(self, obj: AuxObject):
        """Removes `obj` from self."""
        if not isinstance(obj, AuxObject):
            raise TypeError(f"Expected AuxObject, got {obj!r}")
        found = self._key_map.get(obj.key)
        if found is None:
            return
        elif found == obj:
            del self._key_map[obj.key]
            obj._unsub()
        else:
            # FIXME: Object has different key, or another object has same key
            return

    def filter_by_type(self, T: Type[AuxObject]) -> List[AuxObject]:
        """Returns all `AuxObject`s with the given type.

        You can pass in Union types as well.
        """

        if get_origin(T) == Union:
            T_raw = get_args(T)
            T = tuple(x for x in T_raw if x is not None)
            # T is a tuple of types

        def chk(x):
            try:
                return isinstance(x, T)
            except Exception:
                return False

        return [x for x in self._key_map.values() if chk(x)]


class RemoveAuxAction(Action):
    """Removes the aux object, at the end of the stack."""

    def __init__(
        self,
        game: Game,
        source: GameObject,
        /,
        target: AuxObject,
        *,
        priority: float = -100,
        canceled: bool = False,
    ):
        self._target = target
        super().__init__(game, source, priority=priority, canceled=canceled)

    @property
    def target(self) -> AuxObject:
        return self._target

    def doit(self):
        self.game.aux.remove(self.target)



from __future__ import annotations

from typing import Any, List, Optional

from open_mafia_engine.core.all import (
    Ability,
    Action,
    Actor,
    ATBase,
    AuxObject,
    CancelAction,
    EPreAction,
    EPostPhaseChange,
    Game,
    GameObject,
    PhaseChangeAction,
    RemoveAuxAction,
    handler,
)


class TempPhaseAux(AuxObject):
    """Aux object that removes itself after the phase."""

    @handler
    def remove_self(self, event: EPostPhaseChange) -> Optional[List[RemoveAuxAction]]:
        if isinstance(event, EPostPhaseChange):
            return [RemoveAuxAction(self.game, self, target=self)]


class ValueAux(AuxObject):
    """Aux object that represents a key-value pair."""

    def __init__(
        self,
        game: Game,
        /,
        key: str,
        value: Any,
        *,
        use_default_constraints: bool = True,
    ):
        super().__init__(game, key=key, use_default_constraints=use_default_constraints)
        self._value = self._fix_val(value)

    def _fix_val(self, v):
        """Fixes the value. Override this if needed."""
        return v

    @property
    def key(self) -> str:
        return self._key

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, v):
        # TODO: Event when value is changed?
        self._value = self._fix_val(v)

    @classmethod
    def get_or_create(cls, game: Game, /, key: str, value: Any, **kwargs) -> ValueAux:
        return super().get_or_create(game, key=key, value=value, **kwargs)


class CounterAux(ValueAux):
    """Int-valued aux object, with a default value of 0."""

    def __init__(
        self,
        game: Game,
        /,
        key: str,
        value: int = 0,
        *,
        use_default_constraints: bool = True,
    ):
        super().__init__(
            game, key=key, value=value, use_default_constraints=use_default_constraints
        )

    def _fix_val(self, v):
        return int(v)

    def increment(self):
        """Increments this counter by one (e.g. `self.value += 1`)"""
        self.value += 1

    @classmethod
    def get_or_create(
        cls, game: Game, /, key: str, value: int = 0, **kwargs
    ) -> CounterAux:
        return super().get_or_create(game, key=key, value=value, **kwargs)


class CounterPerPhaseAux(CounterAux):
    """CounterAux that resets after every phase ends."""

    @handler
    def reset_every_phase(self, event: EPostPhaseChange) -> None:
        if isinstance(event, EPostPhaseChange):
            self.value = 0



from open_mafia_engine.util.matcher import FuzzyMatcher


def test_fuzzy():
    """Tests fuzzy matching at different levels.

    TODO: Make test more comprehensive. Maybe use real-world cases?
    """

    matcher = FuzzyMatcher[str]({"abacus": "a", "peter": "p"}, score_cutoff=0)
    assert matcher["poodle"] == "p"



from __future__ import annotations

from typing import Callable, Dict, List

from open_mafia_engine.core.game import Game
from open_mafia_engine.errors import MafiaBadBuilder
from open_mafia_engine.util.matcher import FuzzyMatcher
from open_mafia_engine.util.repr import ReprMixin

_GameBuilderFunc = Callable[..., Game]  # This signature is not quite enough.

__builders__: Dict[str, _GameBuilderFunc] = {}


def _assert_game_builder(func: Callable):
    """Raises an exception if not a game builder."""

    if not callable(func):
        raise MafiaBadBuilder(func)

    # TODO: Implement! Inspect the resulting signature?


class GameBuilder(ReprMixin):
    """Specification for a Game Builder.

    TODO: Allow lazy loading of specs, somehow.
    """

    def __init__(self, func: _GameBuilderFunc, /, name: str):
        global __builders__

        _assert_game_builder(func)
        self._func = func
        self._name = name

        __builders__[name] = self

    def build(self, player_names: List[str], *args, **kwargs) -> Game:
        """Builds the game based on passed players and other options."""
        return self.func(player_names, *args, **kwargs)

    @property
    def func(self) -> _GameBuilderFunc:
        return self._func

    @property
    def name(self) -> str:
        return self._name

    @classmethod
    def load(cls, name: str) -> GameBuilder:
        """Load a spec by exact or closest name."""
        global __builders__
        matcher = FuzzyMatcher(__builders__, score_cutoff=10)
        return matcher[name]

    @classmethod
    def available(cls) -> List[str]:
        """Returns the names of available builders."""
        global __builders__
        return list(__builders__.keys())


def game_builder(name: str = None) -> Callable[[_GameBuilderFunc], _GameBuilderFunc]:
    """Decorator factor for game builders."""

    def inner(func: _GameBuilderFunc) -> _GameBuilderFunc:
        nonlocal name

        if name is None:
            name = func.__qualname__
        GameBuilder(func, name=name)

        return func

    return inner



# flake8: noqa

from .auxiliary import (
    CounterAux,
    CounterPerPhaseAux,
    RemoveAuxAction,
    TempPhaseAux,
    ValueAux,
)
from .constraints import (
    ConstraintActorTargetsAlive,
    ConstraintNoSelfFactionTarget,
    ConstraintNoSelfTarget,
    ConstraintOwnerAlive,
    LimitPerPhaseActorConstraint,
    LimitPerPhaseKeyConstraint,
    PhaseConstraint,
)
from .information import (
    BaseInformationAction,
    BaseInspectAction,
    InspectFactionAbility,
    InspectFactionAction,
)
from .kills import DeathCausingAction, KillAbility, KillAction, LynchAction
from .lynch_tally import LynchTally
from .notify import EFactionNotify, EFActorNotify, ENotify
from .outcome import OCLastFactionStanding
from .phases import PhaseChangeAbility
from .protect import ProtectFromKillAbility, ProtectFromKillAction, ProtectorFromKillAux
from .redirect import (
    ActorRedirectAction,
    ActorRedirectorAux,
    BaseRedirectAction,
    CreateRedirectAbility,
    CreateRedirectAction,
)
from .roleblock import RoleBlockAbility, RoleBlockAction, RoleBlockerAux
from .startup import ECreateFactionChat, FactionChatCreatorAux
from .triggers import UnkillableTrigger, UnlynchableTrigger
from .voting import (
    AbstractVoteTarget,
    ActorTarget,
    ActorTargets,
    Tally,
    UnvoteAll,
    Vote,
    VoteAbility,
    VoteAction,
    VoteAgainstAll,
    VotingOptions,
    VotingResults,
    get_vote_target,
    get_vote_target_multi,
    get_vote_target_single,
)
